"""Shared config normalization helpers."""

from __future__ import annotations

from dagzoo.math import normalize_positive_weights

from .constants import (
    _MECHANISM_FAMILY_VALUE_MAP,
    _MISSINGNESS_MECHANISM_VALUE_MAP,
    _NOISE_FAMILY_VALUE_MAP,
    _NOISE_MIXTURE_COMPONENT_VALUE_MAP,
    _SHIFT_MODE_VALUE_MAP,
    MechanismFamily,
    MissingnessMechanism,
    NoiseFamily,
    NoiseMixtureComponent,
    ShiftMode,
)
from .scalars import _validate_finite_float_field


def normalize_missing_mechanism(value: str) -> MissingnessMechanism:
    """Normalize missingness mechanism into a validated internal value."""

    if isinstance(value, bool) or not isinstance(value, str):
        raise ValueError(
            f"Unsupported missing_mechanism '{value}'. Expected none, mcar, mar, or mnar."
        )
    normalized = value.strip().lower()
    result = _MISSINGNESS_MECHANISM_VALUE_MAP.get(normalized)
    if result is None:
        raise ValueError(
            f"Unsupported missing_mechanism '{value}'. Expected none, mcar, mar, or mnar."
        )
    return result


def normalize_shift_mode(value: object) -> ShiftMode:
    """Normalize shift mode into a validated internal value."""

    if isinstance(value, bool) or not isinstance(value, str):
        raise ValueError(
            "Unsupported shift.mode "
            f"'{value}'. Expected off, graph_drift, mechanism_drift, noise_drift, mixed, or custom."
        )
    normalized = value.strip().lower()
    result = _SHIFT_MODE_VALUE_MAP.get(normalized)
    if result is None:
        raise ValueError(
            "Unsupported shift.mode "
            f"'{value}'. Expected off, graph_drift, mechanism_drift, noise_drift, mixed, or custom."
        )
    return result


def normalize_noise_family(value: object) -> NoiseFamily:
    """Normalize noise family into a validated internal value."""

    if isinstance(value, bool) or not isinstance(value, str):
        raise ValueError(
            "Unsupported noise.family "
            f"'{value}'. Expected gaussian, laplace, student_t, or mixture."
        )
    normalized = value.strip().lower()
    result = _NOISE_FAMILY_VALUE_MAP.get(normalized)
    if result is None:
        raise ValueError(
            "Unsupported noise.family "
            f"'{value}'. Expected gaussian, laplace, student_t, or mixture."
        )
    return result


def _normalize_noise_mixture_weights(
    value: object | None,
) -> dict[NoiseMixtureComponent, float] | None:
    """Normalize optional noise-mixture weights mapping."""

    if value is None:
        return None
    if not isinstance(value, dict):
        raise ValueError("noise.mixture_weights must be a mapping.")
    if not value:
        raise ValueError(
            "noise.mixture_weights must include at least one of gaussian, laplace, or student_t."
        )

    weights: dict[NoiseMixtureComponent, float] = {}
    for raw_key, raw_weight in value.items():
        if isinstance(raw_key, bool) or not isinstance(raw_key, str):
            raise ValueError(
                "noise.mixture_weights keys must be gaussian, laplace, or student_t strings."
            )
        normalized_key = raw_key.strip().lower()
        component = _NOISE_MIXTURE_COMPONENT_VALUE_MAP.get(normalized_key)
        if component is None:
            raise ValueError(
                "Unsupported noise.mixture_weights key "
                f"'{raw_key}'. Expected gaussian, laplace, or student_t."
            )
        if component in weights:
            raise ValueError(
                f"Duplicate noise.mixture_weights key '{raw_key}' after normalization."
            )
        weight = _validate_finite_float_field(
            field_name=f"noise.mixture_weights.{component}",
            value=raw_weight,
            lo=0.0,
            hi=None,
            lo_inclusive=True,
            hi_inclusive=False,
            expectation="a finite value >= 0",
        )
        weights[component] = float(weight)

    return normalize_positive_weights(weights, field_name="noise.mixture_weights")


def _normalize_function_family_mix(
    value: object | None,
) -> dict[MechanismFamily, float] | None:
    """Normalize optional mechanism family-mix weights mapping."""

    if value is None:
        return None
    if not isinstance(value, dict):
        raise ValueError("mechanism.function_family_mix must be a mapping.")
    if not value:
        raise ValueError(
            "mechanism.function_family_mix must include at least one supported family."
        )

    weights: dict[MechanismFamily, float] = {}
    for raw_key, raw_weight in value.items():
        if isinstance(raw_key, bool) or not isinstance(raw_key, str):
            raise ValueError("mechanism.function_family_mix keys must be mechanism family names.")
        normalized_key = raw_key.strip().lower()
        family = _MECHANISM_FAMILY_VALUE_MAP.get(normalized_key)
        if family is None:
            supported = ", ".join(sorted(_MECHANISM_FAMILY_VALUE_MAP))
            raise ValueError(
                "Unsupported mechanism.function_family_mix key "
                f"'{raw_key}'. Expected one of: {supported}."
            )
        if family in weights:
            raise ValueError(
                f"Duplicate mechanism.function_family_mix key '{raw_key}' after normalization."
            )
        weight = _validate_finite_float_field(
            field_name=f"mechanism.function_family_mix.{family}",
            value=raw_weight,
            lo=0.0,
            hi=None,
            lo_inclusive=True,
            hi_inclusive=False,
            expectation="a finite value >= 0",
        )
        weights[family] = float(weight)
    return normalize_positive_weights(
        weights,
        field_name="mechanism.function_family_mix",
    )



import pytest
import torch

from dagzoo.filtering import apply_extra_trees_filter
from dagzoo.filtering.extra_trees_filter import _apply_extra_trees_filter_numpy


def _make_regression_data(
    seed: int = 7, n_rows: int = 256, n_features: int = 12
) -> tuple[torch.Tensor, torch.Tensor]:
    g = torch.Generator(device="cpu")
    g.manual_seed(seed)
    x = torch.randn(n_rows, n_features, generator=g)
    w = torch.randn(n_features, 1, generator=g)
    y = (x @ w).squeeze(1) + 0.1 * torch.randn(n_rows, generator=g)
    return x, y


def _make_classification_data(
    seed: int = 11,
    n_rows: int = 256,
    n_features: int = 10,
    n_classes: int = 4,
) -> tuple[torch.Tensor, torch.Tensor]:
    g = torch.Generator(device="cpu")
    g.manual_seed(seed)
    y = torch.arange(n_rows, dtype=torch.int64) % int(n_classes)
    y = y[torch.randperm(n_rows, generator=g)]
    centroids = torch.randn(n_classes, n_features, generator=g)
    x = centroids[y] + 0.25 * torch.randn(n_rows, n_features, generator=g)
    return x, y


def _make_sparse_label_classification_data(
    *,
    seed: int = 91,
    n_rows: int = 256,
    n_features: int = 10,
    labels: tuple[int, ...] = (10, 20, 30, 40),
) -> tuple[torch.Tensor, torch.Tensor]:
    g = torch.Generator(device="cpu")
    g.manual_seed(seed)
    class_ids = torch.arange(n_rows, dtype=torch.int64) % int(len(labels))
    class_ids = class_ids[torch.randperm(n_rows, generator=g)]
    label_values = torch.tensor(labels, dtype=torch.int64)
    y = label_values[class_ids]
    centroids = torch.randn(len(labels), n_features, generator=g)
    x = centroids[class_ids] + 0.25 * torch.randn(n_rows, n_features, generator=g)
    return x, y


def test_extra_trees_filter_is_deterministic_for_fixed_seed() -> None:
    x, y = _make_regression_data()
    accepted_a, details_a = apply_extra_trees_filter(
        x,
        y,
        task="regression",
        seed=123,
        n_estimators=8,
        max_depth=5,
        n_bootstrap=33,
        threshold=0.5,
    )
    accepted_b, details_b = apply_extra_trees_filter(
        x,
        y,
        task="regression",
        seed=123,
        n_estimators=8,
        max_depth=5,
        n_bootstrap=33,
        threshold=0.5,
    )

    assert accepted_a == accepted_b
    assert details_a["wins_ratio"] == details_b["wins_ratio"]
    assert details_a["n_valid_oob"] == details_b["n_valid_oob"]
    assert details_a["backend"] == "extra_trees_cpu"
    assert int(details_a["n_jobs"]) == -1


@pytest.mark.parametrize("task", ["classification", "regression"])
def test_extra_trees_filter_numpy_helper_matches_torch_wrapper(task: str) -> None:
    if task == "classification":
        x, y = _make_classification_data(seed=141)
    else:
        x, y = _make_regression_data(seed=141)

    accepted_torch, details_torch = apply_extra_trees_filter(
        x,
        y,
        task=task,
        seed=123,
        n_estimators=8,
        max_depth=5,
        min_samples_leaf=2,
        n_bootstrap=24,
        threshold=0.5,
        n_jobs=1,
    )
    x_np = x.detach().cpu().numpy()
    y_np = y.detach().cpu().numpy()
    accepted_numpy, details_numpy = _apply_extra_trees_filter_numpy(
        x_np,
        y_np,
        task=task,
        seed=123,
        n_estimators=8,
        max_depth=5,
        min_samples_leaf=2,
        n_bootstrap=24,
        threshold=0.5,
        n_jobs=1,
    )

    assert accepted_numpy == accepted_torch
    assert details_numpy == details_torch


def test_extra_trees_filter_handles_non_divisible_bootstrap_chunks() -> None:
    x, y = _make_regression_data(seed=17)
    accepted, details = apply_extra_trees_filter(
        x,
        y,
        task="regression",
        seed=44,
        n_estimators=6,
        max_depth=4,
        n_bootstrap=33,
        threshold=0.5,
    )

    assert isinstance(accepted, bool)
    assert "wins_ratio" in details
    assert "n_valid_oob" in details
    assert details["backend"] == "extra_trees_cpu"


def test_extra_trees_filter_enforces_max_leaf_nodes() -> None:
    x, y = _make_regression_data(seed=21)
    accepted, details = apply_extra_trees_filter(
        x,
        y,
        task="regression",
        seed=42,
        n_estimators=6,
        max_depth=6,
        max_leaf_nodes=1,
        n_bootstrap=16,
        threshold=0.5,
    )

    assert isinstance(accepted, bool)
    assert details["backend"] == "extra_trees_cpu"


def test_extra_trees_filter_can_report_insufficient_oob_predictions() -> None:
    x, y = _make_regression_data(seed=5, n_rows=16, n_features=8)
    accepted, details = apply_extra_trees_filter(
        x,
        y,
        task="regression",
        seed=99,
        n_estimators=1,
        max_depth=0,
        min_samples_leaf=2,
        n_bootstrap=16,
        threshold=0.5,
    )

    assert accepted is False
    assert details["reason"] == "insufficient_oob_predictions"
    assert 0 <= int(details["n_valid_oob"]) < 16
    assert details["backend"] == "extra_trees_cpu"
    assert int(details["n_jobs"]) == -1
    assert details["threshold_requested"] == pytest.approx(0.5)
    assert details["threshold_effective"] == pytest.approx(0.5)
    assert details["threshold_policy"] == "class_aware_piecewise_v1"
    assert details["class_count"] is None
    assert details["class_bucket"] == "not_applicable"
    assert details["threshold_delta"] == pytest.approx(0.0)


def test_extra_trees_filter_excludes_rows_with_no_oob_votes() -> None:
    n_rows = 64
    x, y = _make_regression_data(seed=13, n_rows=n_rows, n_features=8)
    accepted, details = apply_extra_trees_filter(
        x,
        y,
        task="regression",
        seed=115,
        n_estimators=1,
        max_depth=6,
        min_samples_leaf=1,
        n_bootstrap=8,
        threshold=0.1,
    )

    assert accepted is False
    assert details["reason"] == "insufficient_oob_predictions"
    assert 0 <= int(details["n_valid_oob"]) < n_rows


def test_extra_trees_filter_zero_threshold_bypasses_model_scoring(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    x, y = _make_regression_data(seed=13, n_rows=64, n_features=8)

    def _fail_model_init(*_args, **_kwargs):
        raise AssertionError("ExtraTreesRegressor should not be instantiated in bypass mode")

    monkeypatch.setattr(
        "dagzoo.filtering.extra_trees_filter.ExtraTreesRegressor",
        _fail_model_init,
    )

    accepted, details = apply_extra_trees_filter(
        x,
        y,
        task="regression",
        seed=115,
        n_estimators=1,
        max_depth=6,
        min_samples_leaf=1,
        n_bootstrap=8,
        threshold=0.0,
    )

    assert accepted is True
    assert details["backend"] == "filter_threshold_bypass"
    assert details["bypass"] is True
    assert details["threshold_requested"] == pytest.approx(0.0)
    assert details["threshold_effective"] == pytest.approx(0.0)
    assert details["threshold_policy"] == "zero_bypass_v1"
    assert details["threshold_delta"] == pytest.approx(0.0)
    assert "wins_ratio" not in details
    assert "n_valid_oob" not in details


def test_extra_trees_filter_classification_smoke() -> None:
    x, y = _make_classification_data()
    accepted, details = apply_extra_trees_filter(
        x,
        y,
        task="classification",
        seed=1,
        n_estimators=8,
        max_depth=4,
        min_samples_leaf=2,
        n_bootstrap=24,
        threshold=0.5,
    )

    assert isinstance(accepted, bool)
    assert "wins_ratio" in details
    assert "n_valid_oob" in details
    assert details["backend"] == "extra_trees_cpu"
    assert details["threshold_requested"] == pytest.approx(0.5)
    assert details["threshold_effective"] == pytest.approx(0.5)
    assert details["threshold_policy"] == "class_aware_piecewise_v1"
    assert int(details["class_count"]) >= 2
    assert details["class_bucket"] == "<=8"
    assert details["threshold_delta"] == pytest.approx(0.0)


@pytest.mark.parametrize(
    ("n_classes", "expected_bucket", "expected_effective"),
    [
        (4, "<=8", 0.95),
        (10, "9-16", 0.90),
        (18, "17-24", 0.85),
        (28, "25-32", 0.80),
        (32, "25-32", 0.80),
    ],
)
def test_extra_trees_filter_class_aware_threshold_sweep(
    n_classes: int,
    expected_bucket: str,
    expected_effective: float,
) -> None:
    x, y = _make_classification_data(
        seed=211 + n_classes,
        n_rows=512,
        n_features=12,
        n_classes=n_classes,
    )
    _, details = apply_extra_trees_filter(
        x,
        y,
        task="classification",
        seed=17,
        n_estimators=6,
        max_depth=4,
        min_samples_leaf=2,
        n_bootstrap=16,
        threshold=0.95,
    )

    assert details["class_bucket"] == expected_bucket
    assert details["threshold_requested"] == pytest.approx(0.95)
    assert details["threshold_effective"] == pytest.approx(expected_effective)
    assert details["threshold_policy"] == "class_aware_piecewise_v1"
    assert int(details["class_count"]) == n_classes
    assert details["threshold_delta"] == pytest.approx(0.95 - expected_effective)


def test_extra_trees_filter_regression_threshold_diagnostics_are_not_applicable() -> None:
    x, y = _make_regression_data(seed=123)
    _, details = apply_extra_trees_filter(
        x,
        y,
        task="regression",
        seed=88,
        n_estimators=6,
        max_depth=4,
        n_bootstrap=16,
        threshold=0.75,
    )

    assert details["threshold_requested"] == pytest.approx(0.75)
    assert details["threshold_effective"] == pytest.approx(0.75)
    assert details["threshold_policy"] == "class_aware_piecewise_v1"
    assert details["class_count"] is None
    assert details["class_bucket"] == "not_applicable"
    assert details["threshold_delta"] == pytest.approx(0.0)


def test_extra_trees_filter_rejected_scored_run_keeps_threshold_diagnostics(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    x, y = _make_classification_data(seed=77, n_rows=384, n_features=12, n_classes=32)
    monkeypatch.setattr(
        "dagzoo.filtering.extra_trees_filter._bootstrap_wins_ratio",
        lambda **_kwargs: 0.5,
    )
    accepted, details = apply_extra_trees_filter(
        x,
        y,
        task="classification",
        seed=31,
        n_estimators=6,
        max_depth=4,
        min_samples_leaf=2,
        n_bootstrap=16,
        threshold=1.0,
    )

    assert accepted is False
    assert "wins_ratio" in details
    assert details["threshold_requested"] == pytest.approx(1.0)
    assert details["threshold_effective"] == pytest.approx(0.85)
    assert details["threshold_policy"] == "class_aware_piecewise_v1"
    assert details["class_bucket"] == "25-32"
    assert int(details["class_count"]) == 32
    assert details["threshold_delta"] == pytest.approx(0.15)


def test_extra_trees_filter_many_class_acceptance_rate_is_not_pathological() -> None:
    seeds = [1701, 1702, 1703, 1704, 1705, 1706]
    acceptance_rates: dict[int, float] = {}

    for n_classes in (8, 16, 24, 32):
        accepted = 0
        for seed in seeds:
            x, y = _make_classification_data(
                seed=seed + (n_classes * 100),
                n_rows=512,
                n_features=12,
                n_classes=n_classes,
            )
            is_accepted, _details = apply_extra_trees_filter(
                x,
                y,
                task="classification",
                seed=seed,
                n_estimators=6,
                max_depth=4,
                min_samples_leaf=2,
                n_bootstrap=16,
                threshold=0.95,
            )
            if is_accepted:
                accepted += 1
        acceptance_rates[n_classes] = accepted / float(len(seeds))

    assert acceptance_rates[32] >= (acceptance_rates[8] - 0.25)


@pytest.mark.parametrize(
    ("labels", "expected_class_count", "expected_bucket", "expected_effective"),
    [
        ((10, 11), 2, "<=8", 0.95),
        ((10, 20, 30, 40), 4, "<=8", 0.95),
    ],
)
def test_extra_trees_filter_sparse_labels_use_realized_class_count(
    labels: tuple[int, ...],
    expected_class_count: int,
    expected_bucket: str,
    expected_effective: float,
) -> None:
    x, y = _make_sparse_label_classification_data(labels=labels)
    _, details = apply_extra_trees_filter(
        x,
        y,
        task="classification",
        seed=123,
        n_estimators=6,
        max_depth=4,
        min_samples_leaf=2,
        n_bootstrap=16,
        threshold=0.95,
    )

    assert int(details["class_count"]) == expected_class_count
    assert details["class_bucket"] == expected_bucket
    assert details["threshold_requested"] == pytest.approx(0.95)
    assert details["threshold_effective"] == pytest.approx(expected_effective)
    assert details["threshold_delta"] == pytest.approx(0.95 - expected_effective)


@pytest.mark.parametrize("bad_seed", [-1, 4294967296])
def test_extra_trees_filter_rejects_out_of_range_seed(bad_seed: int) -> None:
    x, y = _make_regression_data(seed=9)
    with pytest.raises(ValueError, match=r"seed must be an integer in \[0, 4294967295\]"):
        apply_extra_trees_filter(
            x,
            y,
            task="regression",
            seed=bad_seed,
            n_estimators=4,
            max_depth=3,
            n_bootstrap=8,
            threshold=0.5,
        )


@pytest.mark.parametrize("seed", [0, 4294967295])
def test_extra_trees_filter_accepts_32bit_seed_boundaries(seed: int) -> None:
    x, y = _make_regression_data(seed=19)
    accepted, details = apply_extra_trees_filter(
        x,
        y,
        task="regression",
        seed=seed,
        n_estimators=4,
        max_depth=3,
        n_bootstrap=8,
        threshold=0.5,
    )
    assert isinstance(accepted, bool)
    assert details["backend"] == "extra_trees_cpu"


def test_extra_trees_filter_honors_explicit_n_jobs() -> None:
    x, y = _make_regression_data(seed=1234)
    accepted, details = apply_extra_trees_filter(
        x,
        y,
        task="regression",
        seed=99,
        n_estimators=4,
        max_depth=3,
        n_bootstrap=8,
        threshold=0.5,
        n_jobs=1,
    )

    assert isinstance(accepted, bool)
    assert int(details["n_jobs"]) == 1


@pytest.mark.parametrize("bad_n_jobs", [0, -2, True])
def test_extra_trees_filter_rejects_invalid_n_jobs(bad_n_jobs: int | bool) -> None:
    x, y = _make_regression_data(seed=123)
    with pytest.raises(ValueError, match=r"n_jobs must be -1 or an integer >= 1"):
        apply_extra_trees_filter(
            x,
            y,
            task="regression",
            seed=42,
            n_estimators=4,
            max_depth=3,
            n_bootstrap=8,
            threshold=0.5,
            n_jobs=bad_n_jobs,
        )


@pytest.mark.parametrize("bad_threshold", [float("nan"), float("inf"), -0.1, 1.1, True])
def test_extra_trees_filter_rejects_invalid_thresholds(bad_threshold: float | bool) -> None:
    x, y = _make_regression_data(seed=123)
    with pytest.raises(ValueError, match=r"threshold must be a finite value in \[0.0, 1.0\]"):
        apply_extra_trees_filter(
            x,
            y,
            task="regression",
            seed=42,
            n_estimators=4,
            max_depth=3,
            n_bootstrap=8,
            threshold=bad_threshold,
        )



import pytest
from conftest import load_repo_config
from hypothesis import given, settings
from hypothesis import strategies as st

from dagzoo.bench.constants import (
    SMOKE_N_FEATURES_CAP,
    SMOKE_N_NODES_CAP,
    SMOKE_N_TEST_CAP,
    SMOKE_N_TRAIN_CAP,
)
from dagzoo.config import (
    DATASET_ROWS_MAX_TOTAL,
    DATASET_ROWS_MIN_TOTAL,
    DatasetRowsSpec,
    GeneratorConfig,
    dataset_rows_bounds,
    normalize_dataset_rows,
)
from dagzoo.core.config_resolution import (
    BenchmarkSmokeCaps,
    cap_rows_spec_to_total,
    resolve_benchmark_preset_config,
    resolve_generate_config,
    serialize_resolution_events,
)
from dagzoo.hardware import HardwareInfo

_ROW_TOTAL_STRATEGY = st.integers(
    min_value=DATASET_ROWS_MIN_TOTAL,
    max_value=DATASET_ROWS_MAX_TOTAL,
)
_TOTAL_ROWS_CAP_STRATEGY = st.integers(
    min_value=DATASET_ROWS_MIN_TOTAL,
    max_value=DATASET_ROWS_MAX_TOTAL,
)


@st.composite
def _cap_rows_spec_strategy(draw: st.DrawFn) -> DatasetRowsSpec:
    mode = draw(st.sampled_from(("fixed", "range", "choices")))
    if mode == "fixed":
        value = draw(_ROW_TOTAL_STRATEGY)
        return DatasetRowsSpec(mode="fixed", value=value)
    if mode == "range":
        start = draw(
            st.integers(
                min_value=DATASET_ROWS_MIN_TOTAL,
                max_value=DATASET_ROWS_MAX_TOTAL - 1,
            )
        )
        stop = draw(st.integers(min_value=start + 1, max_value=DATASET_ROWS_MAX_TOTAL))
        return DatasetRowsSpec(mode="range", start=start, stop=stop)

    choices = draw(st.lists(_ROW_TOTAL_STRATEGY, min_size=2, max_size=5, unique=True))
    return DatasetRowsSpec(mode="choices", choices=choices)


def _mock_cuda_h100(_requested_device: str) -> HardwareInfo:
    return HardwareInfo(
        backend="cuda",
        requested_device="cuda",
        device_name="NVIDIA H100 SXM",
        total_memory_gb=80.0,
        peak_flops=989e12,
        tier="cuda_h100",
    )


def test_resolve_generate_config_applies_cli_overrides() -> None:
    cfg = load_repo_config()
    resolved = resolve_generate_config(
        cfg,
        device_override="cpu",
        rows=None,
        hardware_policy="none",
        missing_rate=0.2,
        missing_mechanism="mnar",
        missing_mar_observed_fraction=0.7,
        missing_mar_logit_scale=1.5,
        missing_mnar_logit_scale=2.5,
        diagnostics_enabled=True,
    )
    assert cfg.runtime.device == "auto"
    assert resolved.requested_device == "cpu"
    assert resolved.config.runtime.device == "cpu"
    assert resolved.config.dataset.missing_rate == pytest.approx(0.2)
    assert resolved.config.dataset.missing_mechanism == "mnar"
    assert resolved.config.dataset.missing_mar_observed_fraction == pytest.approx(0.7)
    assert resolved.config.dataset.missing_mar_logit_scale == pytest.approx(1.5)
    assert resolved.config.dataset.missing_mnar_logit_scale == pytest.approx(2.5)
    assert resolved.config.diagnostics.enabled is True

    trace = serialize_resolution_events(resolved.trace_events)
    assert any(
        event["path"] == "runtime.device" and event["source"] == "cli.device" for event in trace
    )
    assert any(
        event["path"] == "dataset.missing_rate" and event["source"] == "cli.missingness_override"
        for event in trace
    )
    assert any(
        event["path"] == "diagnostics.enabled" and event["source"] == "cli.diagnostics"
        for event in trace
    )


def test_resolve_generate_config_applies_rows_override() -> None:
    cfg = load_repo_config()
    resolved = resolve_generate_config(
        cfg,
        device_override="cpu",
        rows="400..60000",
        hardware_policy="none",
        missing_rate=None,
        missing_mechanism=None,
        missing_mar_observed_fraction=None,
        missing_mar_logit_scale=None,
        missing_mnar_logit_scale=None,
        diagnostics_enabled=False,
    )
    assert resolved.config.dataset.rows is not None
    assert resolved.config.dataset.rows.mode == "range"
    assert resolved.config.dataset.rows.start == 400
    assert resolved.config.dataset.rows.stop == 60000

    trace = serialize_resolution_events(resolved.trace_events)
    assert any(event["path"] == "dataset.rows" and event["source"] == "cli.rows" for event in trace)


def test_cap_rows_spec_to_total_clears_rows_when_cap_is_below_min_total() -> None:
    cfg = load_repo_config()
    cfg.dataset.rows = 1024  # type: ignore[assignment]

    cap_rows_spec_to_total(cfg, total_rows_cap=160)

    assert cfg.dataset.rows is None


@pytest.mark.parametrize(
    ("rows_spec", "expected_mode", "expected_payload"),
    [
        (1024, "fixed", {"value": 500}),
        ("400..60000", "range", {"start": 400, "stop": 500}),
        ("400,600,800", "choices", {"choices": [400, 500]}),
    ],
)
def test_cap_rows_spec_to_total_caps_fixed_range_and_choice_rows(
    rows_spec: object,
    expected_mode: str,
    expected_payload: dict[str, int | list[int]],
) -> None:
    cfg = load_repo_config()
    cfg.dataset.rows = rows_spec  # type: ignore[assignment]

    cap_rows_spec_to_total(cfg, total_rows_cap=500)

    assert cfg.dataset.rows is not None
    assert cfg.dataset.rows.mode == expected_mode
    for field_name, expected_value in expected_payload.items():
        assert getattr(cfg.dataset.rows, field_name) == expected_value


@settings(max_examples=100, deadline=None)
@given(rows_spec=_cap_rows_spec_strategy(), total_rows_cap=_TOTAL_ROWS_CAP_STRATEGY)
def test_cap_rows_spec_to_total_preserves_mode_contracts_hypothesis(
    rows_spec: DatasetRowsSpec,
    total_rows_cap: int,
) -> None:
    cfg = GeneratorConfig.from_dict({})
    cfg.dataset.rows = rows_spec

    cap_rows_spec_to_total(cfg, total_rows_cap=total_rows_cap)

    assert cfg.dataset.rows is not None
    capped_rows = normalize_dataset_rows(cfg.dataset.rows)
    assert capped_rows is not None

    bounds = dataset_rows_bounds(capped_rows)
    assert bounds is not None
    assert bounds[0] <= total_rows_cap
    assert bounds[1] <= total_rows_cap

    if rows_spec.mode == "fixed":
        assert rows_spec.value is not None
        assert capped_rows.mode == "fixed"
        assert capped_rows.value == min(int(rows_spec.value), int(total_rows_cap))
        return

    if rows_spec.mode == "range":
        assert rows_spec.start is not None and rows_spec.stop is not None
        expected_start = min(int(rows_spec.start), int(total_rows_cap))
        expected_stop = min(int(rows_spec.stop), int(total_rows_cap))
        if expected_start == expected_stop:
            assert capped_rows.mode == "fixed"
            assert capped_rows.value == expected_stop
            return

        assert capped_rows.mode == "range"
        assert capped_rows.start == expected_start
        assert capped_rows.stop == expected_stop
        return

    expected_choices = sorted(
        {min(int(choice), int(total_rows_cap)) for choice in rows_spec.choices}
    )
    if len(expected_choices) == 1:
        assert capped_rows.mode == "fixed"
        assert capped_rows.value == expected_choices[0]
        return

    assert capped_rows.mode == "choices"
    assert capped_rows.choices == expected_choices
    assert all(choice <= total_rows_cap for choice in capped_rows.choices)


def test_resolve_generate_config_rejects_invalid_missingness_combination() -> None:
    cfg = load_repo_config()
    with pytest.raises(
        ValueError,
        match="dataset.missing_mechanism must be mcar, mar, or mnar when dataset.missing_rate > 0",
    ):
        resolve_generate_config(
            cfg,
            device_override="cpu",
            rows=None,
            hardware_policy="none",
            missing_rate=0.2,
            missing_mechanism="none",
            missing_mar_observed_fraction=None,
            missing_mar_logit_scale=None,
            missing_mnar_logit_scale=None,
            diagnostics_enabled=False,
        )


def test_resolve_generate_config_treats_null_runtime_device_as_auto() -> None:
    cfg = load_repo_config()
    cfg.runtime.device = None  # type: ignore[assignment]

    resolved = resolve_generate_config(
        cfg,
        device_override=None,
        rows=None,
        hardware_policy="none",
        missing_rate=None,
        missing_mechanism=None,
        missing_mar_observed_fraction=None,
        missing_mar_logit_scale=None,
        missing_mnar_logit_scale=None,
        diagnostics_enabled=False,
    )

    assert resolved.requested_device == "auto"
    assert resolved.config.runtime.device == "auto"


def test_resolve_generate_config_applies_rows_override_after_policy_revalidation(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr("dagzoo.core.config_resolution.detect_hardware", _mock_cuda_h100)
    cfg = load_repo_config()
    cfg.dataset.rows = "400..60000"  # type: ignore[assignment]

    resolved = resolve_generate_config(
        cfg,
        device_override="cuda",
        rows="2000..60000",
        hardware_policy="cuda_tiered_v1",
        missing_rate=None,
        missing_mechanism=None,
        missing_mar_observed_fraction=None,
        missing_mar_logit_scale=None,
        missing_mnar_logit_scale=None,
        diagnostics_enabled=False,
    )

    assert resolved.config.dataset.rows is not None
    assert resolved.config.dataset.rows.mode == "range"
    assert resolved.config.dataset.rows.start == 2000
    assert resolved.config.dataset.rows.stop == 60000
    assert resolved.config.dataset.n_test == 1024
    assert resolved.config.runtime.fixed_layout_target_cells == 240_000_000

    trace = serialize_resolution_events(resolved.trace_events)
    assert any(
        event["path"] == "runtime.fixed_layout_target_cells"
        and event["source"] == "hardware_policy.cuda_tiered_v1"
        and event["new_value"] == 240_000_000
        for event in trace
    )


def test_resolve_generate_config_applies_default_cuda_fixed_layout_floor_without_policy(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr("dagzoo.core.config_resolution.detect_hardware", _mock_cuda_h100)
    cfg = load_repo_config()
    assert cfg.runtime.fixed_layout_target_cells is None

    resolved = resolve_generate_config(
        cfg,
        device_override="cuda",
        rows=None,
        hardware_policy="none",
        missing_rate=None,
        missing_mechanism=None,
        missing_mar_observed_fraction=None,
        missing_mar_logit_scale=None,
        missing_mnar_logit_scale=None,
        diagnostics_enabled=False,
    )

    assert resolved.config.runtime.fixed_layout_target_cells == 240_000_000
    trace = serialize_resolution_events(resolved.trace_events)
    assert any(
        event["path"] == "runtime.fixed_layout_target_cells"
        and event["source"] == "hardware.default_cuda_fixed_layout_target_cells"
        and event["old_value"] is None
        and event["new_value"] == 240_000_000
        for event in trace
    )


def test_resolve_generate_config_preserves_explicit_cuda_target_without_policy(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr("dagzoo.core.config_resolution.detect_hardware", _mock_cuda_h100)
    cfg = load_repo_config()
    cfg.runtime.fixed_layout_target_cells = 32_000_000

    resolved = resolve_generate_config(
        cfg,
        device_override="cuda",
        rows=None,
        hardware_policy="none",
        missing_rate=None,
        missing_mechanism=None,
        missing_mar_observed_fraction=None,
        missing_mar_logit_scale=None,
        missing_mnar_logit_scale=None,
        diagnostics_enabled=False,
    )

    assert resolved.config.runtime.fixed_layout_target_cells == 32_000_000
    trace = serialize_resolution_events(resolved.trace_events)
    assert not any(
        event["path"] == "runtime.fixed_layout_target_cells"
        and event["source"] == "hardware.default_cuda_fixed_layout_target_cells"
        for event in trace
    )


def test_resolve_benchmark_preset_config_applies_smoke_caps() -> None:
    cfg = load_repo_config("benchmark_cpu.yaml")
    assert cfg.dataset.n_train > SMOKE_N_TRAIN_CAP
    assert cfg.dataset.n_test == SMOKE_N_TEST_CAP
    assert cfg.dataset.n_features_max > SMOKE_N_FEATURES_CAP
    assert cfg.graph.n_nodes_max > SMOKE_N_NODES_CAP

    resolved = resolve_benchmark_preset_config(
        preset_key="cpu",
        config=cfg,
        preset_device="cpu",
        suite="smoke",
        hardware_policy="none",
        smoke_caps=BenchmarkSmokeCaps(
            n_train=SMOKE_N_TRAIN_CAP,
            n_test=SMOKE_N_TEST_CAP,
            n_features=SMOKE_N_FEATURES_CAP,
            n_nodes=SMOKE_N_NODES_CAP,
        ),
    )
    assert cfg.dataset.n_train > SMOKE_N_TRAIN_CAP
    assert resolved.config.dataset.n_train == SMOKE_N_TRAIN_CAP
    assert resolved.config.dataset.n_test == SMOKE_N_TEST_CAP
    assert resolved.config.dataset.n_features_max == SMOKE_N_FEATURES_CAP
    assert resolved.config.graph.n_nodes_max == SMOKE_N_NODES_CAP

    trace = serialize_resolution_events(resolved.trace_events)
    assert any(
        event["path"] == "dataset.n_train"
        and event["source"] == "benchmark.suite_smoke_caps"
        and event["new_value"] == SMOKE_N_TRAIN_CAP
        for event in trace
    )


def test_resolve_benchmark_preset_config_revalidates_after_smoke_caps() -> None:
    cfg = GeneratorConfig.from_dict(
        {
            "dataset": {
                "task": "classification",
                "n_train": 32,
                "n_test": 32,
                "n_classes_min": 2,
                "n_classes_max": 16,
            }
        }
    )

    with pytest.raises(
        ValueError,
        match=r"dataset classification split constraints for n_classes_max",
    ):
        resolve_benchmark_preset_config(
            preset_key="cpu",
            config=cfg,
            preset_device="cpu",
            suite="smoke",
            hardware_policy="none",
            smoke_caps=BenchmarkSmokeCaps(
                n_train=8,
                n_test=8,
                n_features=SMOKE_N_FEATURES_CAP,
                n_nodes=SMOKE_N_NODES_CAP,
            ),
        )


def test_resolve_benchmark_preset_config_treats_null_runtime_device_as_auto() -> None:
    cfg = load_repo_config("benchmark_cpu.yaml")
    cfg.runtime.device = None  # type: ignore[assignment]

    resolved = resolve_benchmark_preset_config(
        preset_key="cpu",
        config=cfg,
        preset_device=None,
        suite="standard",
        hardware_policy="none",
        smoke_caps=None,
    )

    assert resolved.requested_device == "auto"
    assert resolved.config.runtime.device == "auto"


def test_resolve_benchmark_preset_config_requires_smoke_caps_for_smoke_suite() -> None:
    cfg = load_repo_config("benchmark_cpu.yaml")
    with pytest.raises(
        ValueError, match="Benchmark smoke suite config resolution requires smoke cap values"
    ):
        resolve_benchmark_preset_config(
            preset_key="cpu",
            config=cfg,
            preset_device="cpu",
            suite="smoke",
            hardware_policy="none",
            smoke_caps=None,
        )


def test_resolve_benchmark_preset_config_preserves_dataset_rows_for_standard_suite() -> None:
    cfg = load_repo_config("benchmark_cpu.yaml")
    cfg.dataset.rows = "2000..60000"  # type: ignore[assignment]

    resolved = resolve_benchmark_preset_config(
        preset_key="cpu",
        config=cfg,
        preset_device="cpu",
        suite="standard",
        hardware_policy="none",
        smoke_caps=None,
    )

    assert resolved.config.dataset.rows is not None
    assert resolved.config.dataset.rows.mode == "range"
    assert resolved.config.dataset.rows.start == 2000
    assert resolved.config.dataset.rows.stop == 60000


def test_resolve_benchmark_preset_config_preserves_dataset_rows_after_policy_transform(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr("dagzoo.core.config_resolution.detect_hardware", _mock_cuda_h100)
    cfg = load_repo_config("benchmark_cpu.yaml")
    cfg.dataset.rows = "2000..60000"  # type: ignore[assignment]

    resolved = resolve_benchmark_preset_config(
        preset_key="cpu",
        config=cfg,
        preset_device="cuda",
        suite="standard",
        hardware_policy="cuda_tiered_v1",
        smoke_caps=None,
    )

    # Benchmark preset resolution now permits variable rows; smoke-suite row
    # capping and one-time run realization happen later in benchmark orchestration.
    assert resolved.config.dataset.rows is not None
    assert resolved.config.dataset.rows.mode == "range"
    assert resolved.config.dataset.rows.start == 2000
    assert resolved.config.dataset.rows.stop == 60000
    assert resolved.requested_device == "cuda"
    assert resolved.config.dataset.n_test == 1024
    assert resolved.config.runtime.fixed_layout_target_cells == 240_000_000


def test_resolve_benchmark_preset_config_preserves_explicit_cuda_budget_without_policy(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr("dagzoo.core.config_resolution.detect_hardware", _mock_cuda_h100)
    cfg = load_repo_config("benchmark_cuda_h100.yaml")
    assert cfg.runtime.fixed_layout_target_cells == 32_000_000

    resolved = resolve_benchmark_preset_config(
        preset_key="cuda_h100",
        config=cfg,
        preset_device=None,
        suite="standard",
        hardware_policy="none",
        smoke_caps=None,
    )

    assert resolved.requested_device == "cuda"
    assert resolved.config.runtime.fixed_layout_target_cells == 32_000_000
    trace = serialize_resolution_events(resolved.trace_events)
    assert not any(
        event["path"] == "runtime.fixed_layout_target_cells"
        and event["source"] == "hardware.default_cuda_fixed_layout_target_cells"
        for event in trace
    )


def test_resolve_benchmark_preset_config_applies_default_cuda_floor_when_unset(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr("dagzoo.core.config_resolution.detect_hardware", _mock_cuda_h100)
    cfg = load_repo_config("benchmark_cuda_h100.yaml")
    cfg.runtime.fixed_layout_target_cells = None

    resolved = resolve_benchmark_preset_config(
        preset_key="cuda_h100",
        config=cfg,
        preset_device=None,
        suite="standard",
        hardware_policy="none",
        smoke_caps=None,
    )

    assert resolved.requested_device == "cuda"
    assert resolved.config.runtime.fixed_layout_target_cells == 240_000_000
    trace = serialize_resolution_events(resolved.trace_events)
    assert any(
        event["path"] == "runtime.fixed_layout_target_cells"
        and event["source"] == "hardware.default_cuda_fixed_layout_target_cells"
        and event["old_value"] is None
        and event["new_value"] == 240_000_000
        for event in trace
    )



"""Generate handoff manifest helpers for downstream corpus consumers."""

from __future__ import annotations

import json
import math
from collections.abc import Mapping
from pathlib import Path
from time import time_ns
from typing import Any, NoReturn, cast

from dagzoo.core.identity import stable_blake2s_hex
from dagzoo.core.staged_artifacts import cleanup_path, promote_staged_path, staged_output_path
from dagzoo.io.lineage_artifact import sha256_hex
from dagzoo.math import sanitize_json

HANDOFF_MANIFEST_FILENAME = "handoff_manifest.json"
GENERATE_HANDOFF_SCHEMA_NAME = "dagzoo_generate_handoff_manifest"
GENERATE_HANDOFF_SCHEMA_VERSION = 1
HANDOFF_SOURCE_FAMILY = "dagzoo.fixed_layout_scm"
HANDOFF_RECOMMENDED_TRAINING_CORPUS = "generated"
HANDOFF_RECOMMENDED_TRAINING_ARTIFACT_KEY = "generated_dir"
HANDOFF_CURATION_POLICY = "none"
_BLAKE2S_HEX_LENGTH = 32
_SHA256_HEX_LENGTH = 64
_GENERATE_OVERRIDE_KEYS = (
    "num_datasets",
    "seed",
    "rows",
    "device",
    "hardware_policy",
    "missing_rate",
    "missing_mechanism",
    "missing_mar_observed_fraction",
    "missing_mar_logit_scale",
    "missing_mnar_logit_scale",
    "diagnostics",
    "diagnostics_out_dir",
    "handoff_root",
)


def _raise(path: str, message: str) -> NoReturn:
    raise ValueError(f"{path}: {message}")


def _require_mapping(value: object, *, path: str) -> Mapping[str, Any]:
    if not isinstance(value, Mapping):
        _raise(path, "must be a mapping")
    return cast(Mapping[str, Any], value)


def _require_non_empty_string(value: object, *, path: str) -> str:
    if not isinstance(value, str) or not value.strip():
        _raise(path, "must be a non-empty string")
    return cast(str, value)


def _require_optional_string(value: object, *, path: str) -> str | None:
    if value is None:
        return None
    return _require_non_empty_string(value, path=path)


def _require_bool(value: object, *, path: str) -> bool:
    if not isinstance(value, bool):
        _raise(path, "must be a boolean")
    return cast(bool, value)


def _require_int(value: object, *, path: str) -> int:
    if isinstance(value, bool) or not isinstance(value, int):
        _raise(path, "must be an integer")
    return int(cast(int, value))


def _require_optional_int(value: object, *, path: str) -> int | None:
    if value is None:
        return None
    return _require_int(value, path=path)


def _require_non_negative_float(value: object, *, path: str) -> float:
    if isinstance(value, bool) or not isinstance(value, (int, float)):
        _raise(path, "must be a finite non-negative number")
    number = float(cast(int | float, value))
    if not math.isfinite(number) or number < 0.0:
        _raise(path, "must be a finite non-negative number")
    return number


def _require_optional_float(value: object, *, path: str) -> float | None:
    if value is None:
        return None
    return _require_non_negative_float(value, path=path)


def _require_hex_string(value: object, *, path: str, expected_length: int) -> str:
    text = _require_non_empty_string(value, path=path)
    if len(text) != expected_length or any(ch not in "0123456789abcdef" for ch in text):
        _raise(path, f"must be a {expected_length}-character lowercase hexadecimal string")
    return text


def _require_rows_override(value: object, *, path: str) -> str | int | None:
    if value is None:
        return None
    if isinstance(value, bool):
        _raise(path, "must be a string, integer, or null")
    if isinstance(value, int):
        return int(value)
    if isinstance(value, str) and value.strip():
        return value
    _raise(path, "must be a string, integer, or null")


def _resolve_path_str(path: str | Path) -> str:
    return str(Path(path).resolve())


def _resolve_optional_path_str(path: str | Path | None) -> str | None:
    if path is None:
        return None
    return _resolve_path_str(path)


def _read_sha256(path: str | Path) -> str:
    return sha256_hex(Path(path).read_bytes())


def _relative_posix_path(path: str | Path, *, start: str | Path) -> str:
    return Path(path).resolve().relative_to(Path(start).resolve()).as_posix()


def _datasets_per_minute(*, datasets: int, elapsed_seconds: float) -> float:
    if elapsed_seconds <= 0.0:
        return 0.0
    return (float(datasets) / float(elapsed_seconds)) * 60.0


def _validate_generate_overrides(overrides: Mapping[str, Any], *, path: str) -> None:
    expected_keys = set(_GENERATE_OVERRIDE_KEYS)
    actual_keys = set(overrides)
    unexpected_keys = sorted(actual_keys - expected_keys)
    if unexpected_keys:
        _raise(path, f"contains unknown keys: {', '.join(unexpected_keys)}")
    missing_keys = sorted(expected_keys - actual_keys)
    if missing_keys:
        _raise(path, f"is missing required keys: {', '.join(missing_keys)}")

    _require_int(overrides.get("num_datasets"), path=f"{path}.num_datasets")
    _require_optional_int(overrides.get("seed"), path=f"{path}.seed")
    _require_rows_override(overrides.get("rows"), path=f"{path}.rows")
    _require_optional_string(overrides.get("device"), path=f"{path}.device")
    _require_non_empty_string(overrides.get("hardware_policy"), path=f"{path}.hardware_policy")
    _require_optional_float(overrides.get("missing_rate"), path=f"{path}.missing_rate")
    _require_optional_string(
        overrides.get("missing_mechanism"),
        path=f"{path}.missing_mechanism",
    )
    _require_optional_float(
        overrides.get("missing_mar_observed_fraction"),
        path=f"{path}.missing_mar_observed_fraction",
    )
    _require_optional_float(
        overrides.get("missing_mar_logit_scale"),
        path=f"{path}.missing_mar_logit_scale",
    )
    _require_optional_float(
        overrides.get("missing_mnar_logit_scale"),
        path=f"{path}.missing_mnar_logit_scale",
    )
    _require_bool(overrides.get("diagnostics"), path=f"{path}.diagnostics")
    _require_optional_string(
        overrides.get("diagnostics_out_dir"),
        path=f"{path}.diagnostics_out_dir",
    )
    _require_non_empty_string(overrides.get("handoff_root"), path=f"{path}.handoff_root")


def _load_generated_identity(
    *,
    generated_dir: str | Path,
    expected_datasets: int,
) -> tuple[str, str]:
    """Read canonical corpus ids from emitted metadata instead of filesystem paths."""

    metadata_paths = sorted(Path(generated_dir).glob("shard_*/metadata.ndjson"))
    if not metadata_paths:
        _raise("generated_dir", "must contain shard metadata under shard_*/metadata.ndjson")

    generate_run_id: str | None = None
    dataset_ids: list[str] = []
    for metadata_path in metadata_paths:
        with metadata_path.open("r", encoding="utf-8") as handle:
            for line_number, raw_line in enumerate(handle, start=1):
                line = raw_line.strip()
                if not line:
                    continue
                try:
                    payload = json.loads(line)
                except json.JSONDecodeError as exc:
                    _raise(
                        f"{metadata_path}:{line_number}",
                        f"contains invalid JSON ({exc.msg})",
                    )
                record = _require_mapping(payload, path=f"{metadata_path}:{line_number}")
                metadata = _require_mapping(
                    record.get("metadata"),
                    path=f"{metadata_path}:{line_number}.metadata",
                )
                split_groups = _require_mapping(
                    metadata.get("split_groups"),
                    path=f"{metadata_path}:{line_number}.metadata.split_groups",
                )
                current_generate_run_id = _require_hex_string(
                    split_groups.get("request_run"),
                    path=f"{metadata_path}:{line_number}.metadata.split_groups.request_run",
                    expected_length=_BLAKE2S_HEX_LENGTH,
                )
                dataset_id = _require_hex_string(
                    metadata.get("dataset_id"),
                    path=f"{metadata_path}:{line_number}.metadata.dataset_id",
                    expected_length=_BLAKE2S_HEX_LENGTH,
                )
                if generate_run_id is None:
                    generate_run_id = current_generate_run_id
                elif current_generate_run_id != generate_run_id:
                    _raise(
                        str(metadata_path),
                        "contains multiple request_run identities; expected one canonical run",
                    )
                dataset_ids.append(dataset_id)

    if generate_run_id is None:
        _raise("generated_dir", "must contain at least one metadata record")

    if len(dataset_ids) != int(expected_datasets):
        _raise(
            "generated_dir",
            "metadata record count does not match generated_datasets "
            f"(records={len(dataset_ids)}, generated_datasets={int(expected_datasets)})",
        )

    generated_corpus_id = stable_blake2s_hex(
        {
            "generate_run_id": generate_run_id,
            "dataset_ids": dataset_ids,
        }
    )
    return generate_run_id, generated_corpus_id


def build_generate_handoff_manifest(
    *,
    config_path: str | Path,
    generate_invocation_overrides: Mapping[str, Any],
    run_root: str | Path,
    generated_dir: str | Path,
    effective_config_path: str | Path,
    effective_config_trace_path: str | Path,
    generated_datasets: int,
    generation_elapsed_seconds: float,
    requested_device: str,
    resolved_device: str,
    hardware_backend: str,
    hardware_device_name: str,
    hardware_tier: str,
    hardware_policy: str,
    diversity_summary_json_path: str | Path | None = None,
    diversity_summary_md_path: str | Path | None = None,
) -> dict[str, Any]:
    """Build the machine-readable generate handoff manifest payload."""

    overrides_payload = dict(generate_invocation_overrides)
    _validate_generate_overrides(overrides_payload, path="generate_invocation.overrides")
    generate_invocation = {
        "config_path": _resolve_path_str(config_path),
        "overrides": overrides_payload,
    }

    effective_config_sha256 = _read_sha256(effective_config_path)
    effective_config_trace_sha256 = _read_sha256(effective_config_trace_path)
    generate_run_id, generated_corpus_id = _load_generated_identity(
        generated_dir=generated_dir,
        expected_datasets=int(generated_datasets),
    )
    payload: dict[str, Any] = {
        "schema_name": GENERATE_HANDOFF_SCHEMA_NAME,
        "schema_version": GENERATE_HANDOFF_SCHEMA_VERSION,
        "identity": {
            "source_family": HANDOFF_SOURCE_FAMILY,
            "generate_run_id": generate_run_id,
            "generated_corpus_id": generated_corpus_id,
        },
        "generate_invocation": generate_invocation,
        "artifacts": {
            "run_root": _resolve_path_str(run_root),
            "generated_dir": _resolve_path_str(generated_dir),
            "effective_config_path": _resolve_path_str(effective_config_path),
            "effective_config_trace_path": _resolve_path_str(effective_config_trace_path),
        },
        "artifacts_relative": {
            "run_root": ".",
            "generated_dir": _relative_posix_path(generated_dir, start=run_root),
            "effective_config_path": _relative_posix_path(effective_config_path, start=run_root),
            "effective_config_trace_path": _relative_posix_path(
                effective_config_trace_path, start=run_root
            ),
        },
        "checksums": {
            "effective_config_sha256": effective_config_sha256,
            "effective_config_trace_sha256": effective_config_trace_sha256,
        },
        "summary": {
            "generated_datasets": int(generated_datasets),
        },
        "throughput": {
            "generation_stage": {
                "generated_datasets": int(generated_datasets),
                "elapsed_seconds": float(generation_elapsed_seconds),
                "datasets_per_minute": _datasets_per_minute(
                    datasets=int(generated_datasets),
                    elapsed_seconds=float(generation_elapsed_seconds),
                ),
            },
        },
        "hardware": {
            "requested_device": str(requested_device),
            "resolved_device": str(resolved_device),
            "backend": str(hardware_backend),
            "device_name": str(hardware_device_name),
            "tier": str(hardware_tier),
            "hardware_policy": str(hardware_policy),
        },
        "diversity_artifacts": {
            "summary_json_path": _resolve_optional_path_str(diversity_summary_json_path),
            "summary_md_path": _resolve_optional_path_str(diversity_summary_md_path),
        },
        "defaults": {
            "recommended_training_corpus": HANDOFF_RECOMMENDED_TRAINING_CORPUS,
            "recommended_training_artifact_key": HANDOFF_RECOMMENDED_TRAINING_ARTIFACT_KEY,
            "curation_policy": HANDOFF_CURATION_POLICY,
        },
    }
    validate_generate_handoff_manifest(payload)
    return payload


def validate_generate_handoff_manifest(payload: Mapping[str, Any]) -> None:
    """Validate the generate handoff manifest wire shape."""

    root = _require_mapping(payload, path="handoff_manifest")
    schema_name = _require_non_empty_string(
        root.get("schema_name"),
        path="handoff_manifest.schema_name",
    )
    if schema_name != GENERATE_HANDOFF_SCHEMA_NAME:
        _raise(
            "handoff_manifest.schema_name",
            f"must equal {GENERATE_HANDOFF_SCHEMA_NAME!r}",
        )
    schema_version = _require_int(
        root.get("schema_version"),
        path="handoff_manifest.schema_version",
    )
    if schema_version != GENERATE_HANDOFF_SCHEMA_VERSION:
        _raise(
            "handoff_manifest.schema_version",
            f"must equal {GENERATE_HANDOFF_SCHEMA_VERSION}",
        )

    identity = _require_mapping(root.get("identity"), path="handoff_manifest.identity")
    source_family = _require_non_empty_string(
        identity.get("source_family"),
        path="handoff_manifest.identity.source_family",
    )
    if source_family != HANDOFF_SOURCE_FAMILY:
        _raise(
            "handoff_manifest.identity.source_family",
            f"must equal {HANDOFF_SOURCE_FAMILY!r}",
        )
    _require_hex_string(
        identity.get("generate_run_id"),
        path="handoff_manifest.identity.generate_run_id",
        expected_length=_BLAKE2S_HEX_LENGTH,
    )
    _require_hex_string(
        identity.get("generated_corpus_id"),
        path="handoff_manifest.identity.generated_corpus_id",
        expected_length=_BLAKE2S_HEX_LENGTH,
    )

    generate_invocation = _require_mapping(
        root.get("generate_invocation"),
        path="handoff_manifest.generate_invocation",
    )
    _require_non_empty_string(
        generate_invocation.get("config_path"),
        path="handoff_manifest.generate_invocation.config_path",
    )
    overrides = _require_mapping(
        generate_invocation.get("overrides"),
        path="handoff_manifest.generate_invocation.overrides",
    )
    _validate_generate_overrides(overrides, path="handoff_manifest.generate_invocation.overrides")

    artifacts = _require_mapping(root.get("artifacts"), path="handoff_manifest.artifacts")
    for key in (
        "run_root",
        "generated_dir",
        "effective_config_path",
        "effective_config_trace_path",
    ):
        _require_non_empty_string(
            artifacts.get(key),
            path=f"handoff_manifest.artifacts.{key}",
        )

    artifacts_relative = _require_mapping(
        root.get("artifacts_relative"),
        path="handoff_manifest.artifacts_relative",
    )
    run_root_relative = _require_non_empty_string(
        artifacts_relative.get("run_root"),
        path="handoff_manifest.artifacts_relative.run_root",
    )
    if run_root_relative != ".":
        _raise("handoff_manifest.artifacts_relative.run_root", "must equal '.'")
    for key in (
        "generated_dir",
        "effective_config_path",
        "effective_config_trace_path",
    ):
        _require_non_empty_string(
            artifacts_relative.get(key),
            path=f"handoff_manifest.artifacts_relative.{key}",
        )

    checksums = _require_mapping(root.get("checksums"), path="handoff_manifest.checksums")
    for key in (
        "effective_config_sha256",
        "effective_config_trace_sha256",
    ):
        _require_hex_string(
            checksums.get(key),
            path=f"handoff_manifest.checksums.{key}",
            expected_length=_SHA256_HEX_LENGTH,
        )

    summary = _require_mapping(root.get("summary"), path="handoff_manifest.summary")
    _require_int(
        summary.get("generated_datasets"), path="handoff_manifest.summary.generated_datasets"
    )

    throughput = _require_mapping(root.get("throughput"), path="handoff_manifest.throughput")
    generation_stage = _require_mapping(
        throughput.get("generation_stage"),
        path="handoff_manifest.throughput.generation_stage",
    )
    _require_int(
        generation_stage.get("generated_datasets"),
        path="handoff_manifest.throughput.generation_stage.generated_datasets",
    )
    _require_non_negative_float(
        generation_stage.get("elapsed_seconds"),
        path="handoff_manifest.throughput.generation_stage.elapsed_seconds",
    )
    _require_non_negative_float(
        generation_stage.get("datasets_per_minute"),
        path="handoff_manifest.throughput.generation_stage.datasets_per_minute",
    )

    hardware = _require_mapping(root.get("hardware"), path="handoff_manifest.hardware")
    for key in (
        "requested_device",
        "resolved_device",
        "backend",
        "device_name",
        "tier",
        "hardware_policy",
    ):
        _require_non_empty_string(
            hardware.get(key),
            path=f"handoff_manifest.hardware.{key}",
        )

    diversity_artifacts = _require_mapping(
        root.get("diversity_artifacts"),
        path="handoff_manifest.diversity_artifacts",
    )
    _require_optional_string(
        diversity_artifacts.get("summary_json_path"),
        path="handoff_manifest.diversity_artifacts.summary_json_path",
    )
    _require_optional_string(
        diversity_artifacts.get("summary_md_path"),
        path="handoff_manifest.diversity_artifacts.summary_md_path",
    )

    defaults = _require_mapping(root.get("defaults"), path="handoff_manifest.defaults")
    recommended_training_corpus = _require_non_empty_string(
        defaults.get("recommended_training_corpus"),
        path="handoff_manifest.defaults.recommended_training_corpus",
    )
    if recommended_training_corpus != HANDOFF_RECOMMENDED_TRAINING_CORPUS:
        _raise(
            "handoff_manifest.defaults.recommended_training_corpus",
            f"must equal {HANDOFF_RECOMMENDED_TRAINING_CORPUS!r}",
        )
    recommended_artifact_key = _require_non_empty_string(
        defaults.get("recommended_training_artifact_key"),
        path="handoff_manifest.defaults.recommended_training_artifact_key",
    )
    if recommended_artifact_key != HANDOFF_RECOMMENDED_TRAINING_ARTIFACT_KEY:
        _raise(
            "handoff_manifest.defaults.recommended_training_artifact_key",
            f"must equal {HANDOFF_RECOMMENDED_TRAINING_ARTIFACT_KEY!r}",
        )
    curation_policy = _require_non_empty_string(
        defaults.get("curation_policy"),
        path="handoff_manifest.defaults.curation_policy",
    )
    if curation_policy != HANDOFF_CURATION_POLICY:
        _raise(
            "handoff_manifest.defaults.curation_policy",
            f"must equal {HANDOFF_CURATION_POLICY!r}",
        )


def write_generate_handoff_manifest(
    *,
    config_path: str | Path,
    generate_invocation_overrides: Mapping[str, Any],
    run_root: str | Path,
    generated_dir: str | Path,
    effective_config_path: str | Path,
    effective_config_trace_path: str | Path,
    generated_datasets: int,
    generation_elapsed_seconds: float,
    requested_device: str,
    resolved_device: str,
    hardware_backend: str,
    hardware_device_name: str,
    hardware_tier: str,
    hardware_policy: str,
    diversity_summary_json_path: str | Path | None = None,
    diversity_summary_md_path: str | Path | None = None,
    out_path: str | Path | None = None,
) -> Path:
    """Write the generate handoff manifest to disk and return its path."""

    manifest_path = (
        Path(out_path) if out_path is not None else Path(run_root) / HANDOFF_MANIFEST_FILENAME
    )
    payload = build_generate_handoff_manifest(
        config_path=config_path,
        generate_invocation_overrides=generate_invocation_overrides,
        run_root=run_root,
        generated_dir=generated_dir,
        effective_config_path=effective_config_path,
        effective_config_trace_path=effective_config_trace_path,
        generated_datasets=generated_datasets,
        generation_elapsed_seconds=generation_elapsed_seconds,
        requested_device=requested_device,
        resolved_device=resolved_device,
        hardware_backend=hardware_backend,
        hardware_device_name=hardware_device_name,
        hardware_tier=hardware_tier,
        hardware_policy=hardware_policy,
        diversity_summary_json_path=diversity_summary_json_path,
        diversity_summary_md_path=diversity_summary_md_path,
    )
    manifest_path.parent.mkdir(parents=True, exist_ok=True)
    staged_manifest_path = staged_output_path(
        parent_dir=manifest_path.parent,
        final_name=manifest_path.name,
        staging_token=str(time_ns()),
    )
    try:
        staged_manifest_path.write_text(
            json.dumps(sanitize_json(payload), indent=2, sort_keys=True, allow_nan=False) + "\n",
            encoding="utf-8",
        )
        promote_staged_path(staged_path=staged_manifest_path, final_path=manifest_path)
    finally:
        cleanup_path(staged_manifest_path)
    return manifest_path


__all__ = [
    "GENERATE_HANDOFF_SCHEMA_NAME",
    "GENERATE_HANDOFF_SCHEMA_VERSION",
    "HANDOFF_MANIFEST_FILENAME",
    "build_generate_handoff_manifest",
    "validate_generate_handoff_manifest",
    "write_generate_handoff_manifest",
]



from __future__ import annotations

import json
from dataclasses import replace

import pytest
import yaml

from dagzoo.cli import main
from dagzoo.config import GeneratorConfig
from dagzoo.diagnostics.coverage import (
    CoverageAggregationConfig,
    CoverageAggregator,
    write_coverage_summary_json,
    write_coverage_summary_markdown,
)
from dagzoo.diagnostics.types import DatasetMetrics


def _metric_fixture(**overrides: float | int | str | None) -> DatasetMetrics:
    base = DatasetMetrics(
        task="classification",
        n_rows=128,
        n_features=16,
        n_classes=3,
        n_categorical_features=4,
        categorical_ratio=0.25,
        graph_edge_density=0.4,
        shift_enabled=0.0,
        shift_graph_scale=0.0,
        shift_mechanism_scale=0.0,
        shift_variance_scale=0.0,
        shift_edge_odds_multiplier=1.0,
        shift_mechanism_nonlinear_mass=0.625,
        shift_noise_variance_multiplier=1.0,
        linearity_proxy=0.5,
        nonlinearity_proxy=0.1,
        wins_ratio_proxy=0.6,
        pearson_abs_mean=0.2,
        pearson_abs_max=0.8,
        spearman_abs_mean=None,
        spearman_abs_max=None,
        class_entropy=1.0,
        majority_minority_ratio=1.2,
        snr_proxy_db=3.0,
        cat_cardinality_min=2,
        cat_cardinality_mean=3.0,
        cat_cardinality_max=5,
    )
    return replace(base, **overrides)


def test_coverage_aggregation_correctness_on_fixtures() -> None:
    agg = CoverageAggregator(
        CoverageAggregationConfig(
            histogram_bins=4,
            quantiles=(0.25, 0.5, 0.75),
            target_bands={"linearity_proxy": (0.0, 1.0)},
        )
    )
    fixtures = [
        _metric_fixture(linearity_proxy=0.1, wins_ratio_proxy=0.4),
        _metric_fixture(linearity_proxy=0.3, wins_ratio_proxy=0.5),
        _metric_fixture(linearity_proxy=0.7, wins_ratio_proxy=0.8),
        _metric_fixture(linearity_proxy=0.9, wins_ratio_proxy=0.95),
    ]
    for payload in fixtures:
        agg.update_metrics(payload)

    summary = agg.build_summary()
    assert summary["num_datasets"] == 4
    lin = summary["metrics"]["linearity_proxy"]
    assert lin["count"] == 4
    assert lin["missing_count"] == 0
    assert lin["observed_min"] == pytest.approx(0.1)
    assert lin["observed_max"] == pytest.approx(0.9)
    assert lin["quantiles"]["p50"] == pytest.approx(0.5)
    assert lin["histogram"]["num_bins"] == 4
    assert "underrepresented_bins" in lin


def test_coverage_artifact_schema_required_keys(tmp_path) -> None:
    agg = CoverageAggregator(CoverageAggregationConfig(histogram_bins=6))
    agg.update_metrics(_metric_fixture())
    summary = agg.build_summary()
    json_path = write_coverage_summary_json(summary, tmp_path / "coverage_summary.json")
    md_path = write_coverage_summary_markdown(summary, tmp_path / "coverage_summary.md")

    assert json_path.exists()
    assert md_path.exists()
    payload = json.loads(json_path.read_text(encoding="utf-8"))
    assert "generated_at" in payload
    assert payload["num_datasets"] == 1
    assert "metrics" in payload

    required_metric_keys = {
        "count",
        "missing_count",
        "observed_min",
        "observed_max",
        "mean",
        "std",
        "sampled_count",
        "sampled_fraction",
        "quantiles",
        "histogram",
        "underrepresented_bins",
        "target_band",
    }
    line_metric = payload["metrics"]["linearity_proxy"]
    assert required_metric_keys.issubset(set(line_metric))
    assert {"num_bins", "covered_bins", "coverage_ratio", "bins"}.issubset(
        set(line_metric["histogram"])
    )


def test_coverage_aggregation_bounds_sample_memory() -> None:
    agg = CoverageAggregator(
        CoverageAggregationConfig(
            histogram_bins=4,
            max_values_per_metric=3,
        )
    )
    for idx in range(40):
        agg.update_metrics(_metric_fixture(linearity_proxy=float(idx) / 40.0))

    summary = agg.build_summary()
    line = summary["metrics"]["linearity_proxy"]
    assert line["count"] == 40
    assert line["sampled_count"] == 3
    assert line["sampled_fraction"] == pytest.approx(3 / 40)
    assert summary["max_values_per_metric"] == 3


def test_underrepresented_regime_detection_against_target_band() -> None:
    agg = CoverageAggregator(
        CoverageAggregationConfig(
            histogram_bins=4,
            underrepresented_threshold=0.5,
            target_bands={"linearity_proxy": (0.0, 1.0)},
        )
    )
    for _ in range(8):
        agg.update_metrics(_metric_fixture(linearity_proxy=0.0))
    for _ in range(2):
        agg.update_metrics(_metric_fixture(linearity_proxy=1.0))

    summary = agg.build_summary()
    under_bins = summary["metrics"]["linearity_proxy"]["underrepresented_bins"]
    assert len(under_bins) >= 2
    assert summary["metrics"]["linearity_proxy"]["target_band"]["in_target_count"] == 10


def test_target_band_counts_do_not_inflate_for_partial_bin_overlap() -> None:
    agg = CoverageAggregator(
        CoverageAggregationConfig(
            histogram_bins=2,
            target_bands={"linearity_proxy": (0.95, 1.05)},
        )
    )
    agg.update_metrics(_metric_fixture(linearity_proxy=0.9))
    agg.update_metrics(_metric_fixture(linearity_proxy=1.1))

    summary = agg.build_summary()
    line = summary["metrics"]["linearity_proxy"]
    assert line["target_band"]["in_target_count"] == 0
    assert line["target_band"]["in_target_fraction"] == pytest.approx(0.0)


def test_target_band_histogram_uses_target_range_for_coverage() -> None:
    agg = CoverageAggregator(
        CoverageAggregationConfig(
            histogram_bins=10,
            underrepresented_threshold=0.5,
            target_bands={"linearity_proxy": (0.0, 1.0)},
        )
    )
    for value in (0.95, 0.96, 0.98, 0.99, 1.0):
        agg.update_metrics(_metric_fixture(linearity_proxy=value))

    summary = agg.build_summary()
    line = summary["metrics"]["linearity_proxy"]
    # Coverage should reflect sparse occupancy inside the full target range.
    assert line["histogram"]["coverage_ratio"] < 1.0
    assert len(line["underrepresented_bins"]) > 0


def test_generate_no_write_with_coverage_enabled_emits_artifacts(
    tmp_path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    def _stub_filter(*_args, **_kwargs):
        return True, {"wins_ratio": 0.8, "n_valid_oob": 64, "backend": "extra_trees_cpu"}

    monkeypatch.setattr("dagzoo.core.metrics_torch.apply_extra_trees_filter", _stub_filter)
    cfg = GeneratorConfig.from_yaml("configs/default.yaml")
    cfg.runtime.device = "cpu"
    cfg.dataset.task = "regression"
    cfg.dataset.n_train = 32
    cfg.dataset.n_test = 16
    cfg.dataset.n_features_min = 8
    cfg.dataset.n_features_max = 8
    cfg.graph.n_nodes_min = 2
    cfg.graph.n_nodes_max = 6
    cfg.output.out_dir = str(tmp_path / "run")
    cfg.diagnostics.enabled = True
    cfg.diagnostics.histogram_bins = 8
    cfg.diagnostics.meta_feature_targets = {"linearity_proxy": [0.2, 0.8]}
    config_path = tmp_path / "coverage_enabled.yaml"
    config_path.write_text(yaml.safe_dump(cfg.to_dict()), encoding="utf-8")

    code = main(
        [
            "generate",
            "--config",
            str(config_path),
            "--num-datasets",
            "1",
            "--device",
            "cpu",
            "--hardware-policy",
            "none",
            "--no-dataset-write",
        ]
    )
    assert code == 0

    json_path = tmp_path / "run" / "coverage_summary.json"
    md_path = tmp_path / "run" / "coverage_summary.md"
    assert json_path.exists()
    assert md_path.exists()
    payload = json.loads(json_path.read_text(encoding="utf-8"))
    assert payload["num_datasets"] == 1
    assert "mechanism_family_summary" in payload
    assert "linearity_proxy" in payload["metrics"]
    assert "shift_edge_odds_multiplier" in payload["metrics"]
    mechanism_summary = payload["mechanism_family_summary"]
    assert "metadata_coverage_rate" in mechanism_summary
    assert "sampled_family_counts" in mechanism_summary
    assert "sampled_variant_counts" in mechanism_summary
    assert "dataset_presence_rate_by_variant" in mechanism_summary
    markdown = md_path.read_text(encoding="utf-8")
    assert "## Mechanism Families" in markdown



#!/usr/bin/env python3
"""Dry-run or remove ignored local artifact trees.

Targets are limited to known ignored runtime/docs outputs under the repo root.
Use ``--apply`` to actually delete them.
"""

from __future__ import annotations

import argparse
import shutil
import sys
from collections.abc import Iterable
from pathlib import Path

REPO_ROOT = Path(__file__).resolve().parents[1]

TARGET_GROUPS: dict[str, tuple[Path, ...]] = {
    "runtime": (
        REPO_ROOT / "data",
        REPO_ROOT / "benchmarks" / "results",
        REPO_ROOT / "effective_config_artifacts",
    ),
    "docs": (
        REPO_ROOT / "public",
        REPO_ROOT / "site" / "public",
        REPO_ROOT / "site" / ".generated",
    ),
}


def _iter_targets(group: str) -> list[Path]:
    if group == "all":
        paths = TARGET_GROUPS["runtime"] + TARGET_GROUPS["docs"]
    else:
        paths = TARGET_GROUPS[group]
    return list(paths)


def _relpath(path: Path, root: Path) -> str:
    try:
        return str(path.relative_to(root))
    except ValueError:
        return str(path)


def _remove_path(path: Path) -> None:
    if path.is_dir():
        shutil.rmtree(path)
    elif path.exists():
        path.unlink()


def parse_args(argv: Iterable[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--group",
        choices=("runtime", "docs", "all"),
        default="all",
        help="Artifact group to inspect or remove.",
    )
    parser.add_argument(
        "--apply",
        action="store_true",
        help="Actually delete the listed paths. Default is dry-run only.",
    )
    return parser.parse_args(list(argv))


def main(argv: Iterable[str] | None = None) -> int:
    args = parse_args(sys.argv[1:] if argv is None else argv)
    targets = _iter_targets(str(args.group))

    found = 0
    for path in targets:
        rel = _relpath(path, REPO_ROOT)
        if not path.exists():
            print(f"Skip missing: {rel}")
            continue
        found += 1
        if args.apply:
            _remove_path(path)
            print(f"Removed: {rel}")
        else:
            print(f"Would remove: {rel}")

    if not args.apply:
        print("Dry run only. Re-run with --apply to remove the listed paths.")
    elif found == 0:
        print("Nothing to remove.")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""H100 benchmark validation runner with host-level GPU telemetry."""

from __future__ import annotations

import argparse
import datetime as dt
import json
import subprocess
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import torch
import yaml

from dagzoo.hardware import detect_hardware

from .gpu_telemetry import NvidiaSmiSampler, summarize_gpu_telemetry, write_gpu_telemetry_csv

SATURATION_TARGET_CELLS = (160_000_000, 240_000_000, 256_000_000)
FEATURE_RUNS: tuple[tuple[str, str], ...] = (
    ("filter_smoke", "configs/preset_filter_benchmark_smoke.yaml"),
    ("missingness_mar_smoke", "configs/preset_missingness_mar.yaml"),
    ("shift_smoke", "configs/preset_shift_benchmark_smoke.yaml"),
    ("noise_smoke", "configs/preset_noise_benchmark_smoke.yaml"),
    ("many_class_smoke", "configs/preset_many_class_benchmark_smoke.yaml"),
    ("mechanism_baseline_smoke", "configs/preset_mechanism_baseline_benchmark_smoke.yaml"),
    ("mechanism_gp_smoke", "configs/preset_mechanism_gp_benchmark_smoke.yaml"),
    ("mechanism_piecewise_smoke", "configs/preset_mechanism_piecewise_benchmark_smoke.yaml"),
)


@dataclass(slots=True)
class ValidationPhase:
    """One benchmark validation phase."""

    name: str
    command: list[str]
    out_dir: Path
    require_telemetry: bool = False
    group: str = "benchmark"
    target_cells: int | None = None


def _repo_root() -> Path:
    return Path(__file__).resolve().parents[3]


def _write_json(payload: dict[str, Any], path: str | Path) -> Path:
    out_path = Path(path)
    out_path.parent.mkdir(parents=True, exist_ok=True)
    with out_path.open("w", encoding="utf-8") as f:
        json.dump(payload, f, indent=2, sort_keys=True)
    return out_path


def _load_json(path: str | Path) -> dict[str, Any] | None:
    json_path = Path(path)
    if not json_path.exists():
        return None
    with json_path.open("r", encoding="utf-8") as f:
        payload = json.load(f)
    return payload if isinstance(payload, dict) else None


def _torch_cuda_report() -> dict[str, Any]:
    return {
        "torch_version": str(torch.__version__),
        "cuda_available": bool(torch.cuda.is_available()),
        "cuda_count": int(torch.cuda.device_count()),
        "device_names": [
            str(torch.cuda.get_device_name(i)) for i in range(int(torch.cuda.device_count()))
        ]
        if torch.cuda.is_available()
        else [],
    }


def _hardware_report() -> dict[str, Any]:
    hardware = detect_hardware("cuda")
    return {
        "backend": str(hardware.backend),
        "requested_device": str(hardware.requested_device),
        "device_name": str(hardware.device_name),
        "total_memory_gb": hardware.total_memory_gb,
        "peak_flops": hardware.peak_flops,
        "tier": str(hardware.tier),
    }


def _build_benchmark_command(
    python_executable: str,
    *,
    extra_args: list[str],
) -> list[str]:
    return [python_executable, "-m", "dagzoo", "benchmark", *extra_args]


def _write_saturation_config(
    *,
    base_config_path: Path,
    out_path: Path,
    target_cells: int,
) -> Path:
    with base_config_path.open("r", encoding="utf-8") as f:
        payload = yaml.safe_load(f)
    if not isinstance(payload, dict):
        raise ValueError(f"Unexpected config payload: {base_config_path}")
    runtime_payload = payload.setdefault("runtime", {})
    if not isinstance(runtime_payload, dict):
        raise ValueError(f"runtime must be a mapping in {base_config_path}")
    runtime_payload["fixed_layout_target_cells"] = int(target_cells)
    with out_path.open("w", encoding="utf-8") as f:
        yaml.safe_dump(payload, f, sort_keys=False, default_flow_style=False)
    return out_path


def _build_validation_phases(out_root: Path, *, python_executable: str) -> list[ValidationPhase]:
    config_root = out_root / "generated_configs"
    config_root.mkdir(parents=True, exist_ok=True)
    repo_root = _repo_root()
    baseline_path = out_root / "baselines" / "cuda_h100_standard.json"
    saturation_base = repo_root / "configs" / "benchmark_cuda_h100_saturation.yaml"

    phases: list[ValidationPhase] = [
        ValidationPhase(
            name="cuda_h100_smoke",
            out_dir=out_root / "cuda_h100_smoke",
            command=_build_benchmark_command(
                python_executable,
                extra_args=[
                    "--preset",
                    "cuda_h100",
                    "--suite",
                    "smoke",
                    "--hardware-policy",
                    "cuda_tiered_v1",
                    "--out-dir",
                    str(out_root / "cuda_h100_smoke"),
                ],
            ),
        ),
        ValidationPhase(
            name="cuda_h100_standard",
            out_dir=out_root / "cuda_h100_standard",
            require_telemetry=True,
            command=_build_benchmark_command(
                python_executable,
                extra_args=[
                    "--preset",
                    "cuda_h100",
                    "--suite",
                    "standard",
                    "--hardware-policy",
                    "cuda_tiered_v1",
                    "--out-dir",
                    str(out_root / "cuda_h100_standard"),
                    "--save-baseline",
                    str(baseline_path),
                ],
            ),
        ),
        ValidationPhase(
            name="cuda_h100_large_shape",
            out_dir=out_root / "cuda_h100_large_shape",
            require_telemetry=True,
            command=_build_benchmark_command(
                python_executable,
                extra_args=[
                    "--config",
                    str(repo_root / "configs" / "benchmark_cuda_h100_large_shape.yaml"),
                    "--preset",
                    "custom",
                    "--suite",
                    "standard",
                    "--hardware-policy",
                    "cuda_tiered_v1",
                    "--out-dir",
                    str(out_root / "cuda_h100_large_shape"),
                ],
            ),
        ),
    ]

    for target_cells in SATURATION_TARGET_CELLS:
        config_path = _write_saturation_config(
            base_config_path=saturation_base,
            out_path=config_root / f"benchmark_cuda_h100_saturation_{target_cells}.yaml",
            target_cells=target_cells,
        )
        phase_name = f"cuda_h100_saturation_{target_cells}"
        out_dir = out_root / "cuda_h100_saturation" / str(target_cells)
        phases.append(
            ValidationPhase(
                name=phase_name,
                out_dir=out_dir,
                require_telemetry=True,
                group="saturation",
                target_cells=int(target_cells),
                command=_build_benchmark_command(
                    python_executable,
                    extra_args=[
                        "--config",
                        str(config_path),
                        "--preset",
                        "custom",
                        "--suite",
                        "standard",
                        "--hardware-policy",
                        "none",
                        "--out-dir",
                        str(out_dir),
                    ],
                ),
            )
        )

    for run_name, feature_config_path in FEATURE_RUNS:
        phases.append(
            ValidationPhase(
                name=run_name,
                out_dir=out_root / run_name,
                command=_build_benchmark_command(
                    python_executable,
                    extra_args=[
                        "--config",
                        str(repo_root / feature_config_path),
                        "--preset",
                        "custom",
                        "--device",
                        "cuda",
                        "--suite",
                        "smoke",
                        "--hardware-policy",
                        "none",
                        "--no-memory",
                        "--out-dir",
                        str(out_root / run_name),
                    ],
                ),
            )
        )
    return phases


def _phase_metrics_from_summary(summary: dict[str, Any] | None) -> dict[str, Any]:
    if not isinstance(summary, dict):
        return {
            "regression_status": None,
            "datasets_per_minute": None,
            "latency_p95_ms": None,
        }
    preset_results = summary.get("preset_results")
    result = preset_results[0] if isinstance(preset_results, list) and preset_results else {}
    regression = summary.get("regression", {})
    datasets_per_minute = None
    latency_p95_ms = None
    if isinstance(result, dict):
        raw_datasets_per_minute = result.get("datasets_per_minute")
        raw_latency_p95_ms = result.get("latency_p95_ms")
        if isinstance(raw_datasets_per_minute, (int, float)):
            datasets_per_minute = float(raw_datasets_per_minute)
        if isinstance(raw_latency_p95_ms, (int, float)):
            latency_p95_ms = float(raw_latency_p95_ms)
    return {
        "regression_status": (
            str(regression.get("status")) if isinstance(regression, dict) else None
        ),
        "datasets_per_minute": datasets_per_minute,
        "latency_p95_ms": latency_p95_ms,
    }


def _phase_validation_status(
    *,
    exit_code: int,
    summary: dict[str, Any] | None,
    require_telemetry: bool,
    telemetry_summary: dict[str, Any] | None,
) -> str:
    if int(exit_code) != 0:
        return "fail"
    if summary is None:
        return "incomplete"
    if require_telemetry:
        telemetry_available = bool(
            isinstance(telemetry_summary, dict) and telemetry_summary.get("telemetry_available")
        )
        if not telemetry_available:
            return "incomplete"
    return "pass"


def _run_validation_phase(
    phase: ValidationPhase,
    *,
    repo_root: Path,
    telemetry_interval_seconds: float,
) -> dict[str, Any]:
    print(f"Running validation phase: {phase.name}")
    phase.out_dir.mkdir(parents=True, exist_ok=True)
    telemetry_csv_path = phase.out_dir / "gpu_telemetry.csv"
    telemetry_summary_path = phase.out_dir / "gpu_telemetry_summary.json"
    summary_path = phase.out_dir / "summary.json"

    sampler: NvidiaSmiSampler | None = None
    if phase.require_telemetry:
        sampler = NvidiaSmiSampler(interval_seconds=telemetry_interval_seconds)
        sampler.start()

    try:
        completed = subprocess.run(phase.command, cwd=repo_root, check=False)
        exit_code = int(completed.returncode)
    finally:
        if sampler is not None:
            sampler.stop()

    summary = _load_json(summary_path)
    telemetry_summary: dict[str, Any] | None = None
    if phase.require_telemetry:
        samples = list(sampler.samples if sampler is not None else [])
        write_gpu_telemetry_csv(samples, telemetry_csv_path)
        telemetry_summary = summarize_gpu_telemetry(samples)
        telemetry_summary["errors"] = list(sampler.errors if sampler is not None else [])
        _write_json(telemetry_summary, telemetry_summary_path)

    metrics = _phase_metrics_from_summary(summary)
    validation_status = _phase_validation_status(
        exit_code=exit_code,
        summary=summary,
        require_telemetry=phase.require_telemetry,
        telemetry_summary=telemetry_summary,
    )
    return {
        "name": phase.name,
        "group": phase.group,
        "target_cells": phase.target_cells,
        "command": [str(part) for part in phase.command],
        "exit_code": int(exit_code),
        "validation_status": validation_status,
        "require_telemetry": bool(phase.require_telemetry),
        "artifacts": {
            "out_dir": str(phase.out_dir.resolve()),
            "summary_json": str(summary_path.resolve()) if summary_path.exists() else None,
            "summary_md": str((phase.out_dir / "summary.md").resolve())
            if (phase.out_dir / "summary.md").exists()
            else None,
            "gpu_telemetry_csv": str(telemetry_csv_path.resolve())
            if phase.require_telemetry
            else None,
            "gpu_telemetry_summary_json": str(telemetry_summary_path.resolve())
            if phase.require_telemetry
            else None,
        },
        **metrics,
    }


def _build_saturation_summary(phase_results: list[dict[str, Any]]) -> dict[str, Any]:
    saturation_results = [result for result in phase_results if result.get("group") == "saturation"]
    ranked_candidates = [
        result
        for result in saturation_results
        if isinstance(result.get("datasets_per_minute"), (int, float))
    ]
    if not ranked_candidates:
        return {
            "candidate_count": len(saturation_results),
            "recommended_fixed_layout_target_cells": None,
            "recommendation_source": "datasets_per_minute desc, target_cells asc",
            "candidates": saturation_results,
        }

    best = max(
        ranked_candidates,
        key=lambda result: (
            float(result.get("datasets_per_minute", 0.0)),
            -int(result.get("target_cells", 0) or 0),
        ),
    )
    return {
        "candidate_count": len(saturation_results),
        "recommended_fixed_layout_target_cells": int(best["target_cells"]),
        "recommendation_source": "datasets_per_minute desc, target_cells asc",
        "candidates": saturation_results,
    }


def run_h100_validation(
    *,
    out_root: str | Path,
    telemetry_interval_seconds: float = 1.0,
) -> dict[str, Any]:
    """Run the bounded H100 validation suite and return the root manifest."""

    out_dir = Path(out_root)
    out_dir.mkdir(parents=True, exist_ok=True)
    repo_root = _repo_root()
    torch_report = _torch_cuda_report()
    hardware_report = _hardware_report()

    manifest: dict[str, Any] = {
        "generated_at": dt.datetime.now(dt.UTC).isoformat(),
        "out_root": str(out_dir.resolve()),
        "torch_cuda": torch_report,
        "hardware": hardware_report,
    }

    if not torch_report["cuda_available"]:
        manifest["overall_status"] = "fail"
        manifest["phases"] = []
        _write_json(manifest, out_dir / "validation_manifest.json")
        return manifest
    if hardware_report.get("backend") != "cuda":
        manifest["overall_status"] = "fail"
        manifest["phases"] = []
        _write_json(manifest, out_dir / "validation_manifest.json")
        return manifest

    phases = _build_validation_phases(out_dir, python_executable=sys.executable)
    phase_results = [
        _run_validation_phase(
            phase,
            repo_root=repo_root,
            telemetry_interval_seconds=telemetry_interval_seconds,
        )
        for phase in phases
    ]
    manifest["phases"] = phase_results
    manifest["saturation"] = _build_saturation_summary(phase_results)

    overall_status = "pass"
    if any(result.get("validation_status") == "fail" for result in phase_results):
        overall_status = "fail"
    elif any(result.get("validation_status") == "incomplete" for result in phase_results):
        overall_status = "incomplete"
    manifest["overall_status"] = overall_status
    _write_json(manifest, out_dir / "validation_manifest.json")
    return manifest


def _build_arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="python -m dagzoo.bench.h100_validation",
        description="Run the bounded H100 benchmark validation workflow.",
    )
    parser.add_argument(
        "--out-root",
        default=str(Path("benchmarks") / "results" / "gpu_h100_validation"),
        help="Artifact root for benchmark summaries, telemetry, and the validation manifest.",
    )
    parser.add_argument(
        "--telemetry-interval-seconds",
        type=float,
        default=1.0,
        help="Polling interval for host-level nvidia-smi sampling.",
    )
    return parser


def main(argv: list[str] | None = None) -> int:
    """CLI entrypoint for the H100 validation runner."""

    parser = _build_arg_parser()
    args = parser.parse_args(argv)
    manifest = run_h100_validation(
        out_root=args.out_root,
        telemetry_interval_seconds=float(args.telemetry_interval_seconds),
    )
    print(f"Validation manifest: {Path(args.out_root).resolve() / 'validation_manifest.json'}")
    return 0 if manifest.get("overall_status") == "pass" else 1


if __name__ == "__main__":  # pragma: no cover
    raise SystemExit(main())



#!/usr/bin/env python3
"""Validate that a pull request version bump is unchanged or a single patch/minor step."""

from __future__ import annotations

import argparse
import subprocess
import sys
from dataclasses import dataclass
from pathlib import Path

SCRIPT_DIR = Path(__file__).resolve().parents[1]
if str(SCRIPT_DIR) not in sys.path:
    sys.path.insert(0, str(SCRIPT_DIR))

from devlib.semver import (  # noqa: E402
    SemVer,
    allowed_pr_successors,
    read_pyproject_version,
    read_pyproject_version_text,
)


@dataclass(frozen=True)
class PRVersionDecision:
    ok: bool
    head_version: str
    base_version: str
    reason: str


def read_base_pyproject_version(*, base_ref: str, pyproject_path: str) -> str:
    repo_path = Path(pyproject_path).as_posix()
    result = subprocess.run(
        ["git", "show", f"{base_ref}:{repo_path}"],
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode != 0:
        stderr = result.stderr.strip()
        raise ValueError(
            f"Could not read {repo_path} from {base_ref}. {stderr or 'Ensure the base ref exists.'}"
        )
    return read_pyproject_version_text(result.stdout, source=f"{base_ref}:{repo_path}")


def resolve_pr_version_decision(*, head_version: str, base_version: str) -> PRVersionDecision:
    head = SemVer.parse(head_version)
    base = SemVer.parse(base_version)
    if head == base:
        return PRVersionDecision(
            ok=True,
            head_version=str(head),
            base_version=str(base),
            reason="unchanged",
        )

    allowed = allowed_pr_successors(base)
    if head in allowed:
        reason = (
            "next_patch" if head.major == base.major and head.minor == base.minor else "next_minor"
        )
        return PRVersionDecision(
            ok=True,
            head_version=str(head),
            base_version=str(base),
            reason=reason,
        )

    allowed_list = ", ".join(str(version) for version in sorted(allowed))
    raise ValueError(
        f"Pull request version {head} must be unchanged or one allowed step after {base}. "
        f"Allowed bumped versions: {allowed_list}."
    )


def main(argv: list[str] | None = None) -> int:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--pyproject", default="pyproject.toml")
    parser.add_argument("--base-ref", default="origin/main")
    args = parser.parse_args(argv)

    head_version = read_pyproject_version(args.pyproject)
    base_version = read_base_pyproject_version(
        base_ref=args.base_ref, pyproject_path=args.pyproject
    )
    decision = resolve_pr_version_decision(
        head_version=head_version,
        base_version=base_version,
    )
    print(
        f"PR version check passed: head={decision.head_version} "
        f"base={decision.base_version} reason={decision.reason}"
    )
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""Argument parser types and shared CLI choices."""

from __future__ import annotations

import argparse
import math

from dagzoo.config import (
    MISSINGNESS_MECHANISM_MAR,
    MISSINGNESS_MECHANISM_MCAR,
    MISSINGNESS_MECHANISM_MNAR,
    MISSINGNESS_MECHANISM_NONE,
    normalize_missing_mechanism,
)
from dagzoo.filter_thresholds import (
    FILTER_THRESHOLD_EXPECTATION,
    FILTER_THRESHOLD_MAX,
    FILTER_THRESHOLD_MIN,
)
from dagzoo.hardware_policy import list_hardware_policies
from dagzoo.rng import SEED32_MAX, SEED32_MIN

DEVICE_CHOICES = ("auto", "cpu", "cuda", "mps")
HARDWARE_POLICY_CHOICES = list_hardware_policies()
MISSINGNESS_MECHANISM_CLI_CHOICES = (
    MISSINGNESS_MECHANISM_NONE,
    MISSINGNESS_MECHANISM_MCAR,
    MISSINGNESS_MECHANISM_MAR,
    MISSINGNESS_MECHANISM_MNAR,
)


def positive_int(value: str) -> int:
    """argparse type: parse an integer > 0."""

    parsed = int(value)
    if parsed <= 0:
        raise argparse.ArgumentTypeError(f"Expected a positive integer, got {value}.")
    return parsed


def non_negative_int(value: str) -> int:
    """argparse type: parse an integer >= 0."""

    parsed = int(value)
    if parsed < 0:
        raise argparse.ArgumentTypeError(f"Expected a non-negative integer, got {value}.")
    return parsed


def seed_32bit_int(value: str) -> int:
    """argparse type: parse an integer seed in the unsigned 32-bit range."""

    parsed = int(value)
    if parsed < SEED32_MIN or parsed > SEED32_MAX:
        raise argparse.ArgumentTypeError(
            f"Expected a seed in [{SEED32_MIN}, {SEED32_MAX}], got {value}."
        )
    return parsed


def filter_n_jobs(value: str) -> int:
    """argparse type: parse filter worker count (-1 or >= 1)."""

    parsed = int(value)
    if parsed == -1 or parsed >= 1:
        return parsed
    raise argparse.ArgumentTypeError(f"Expected -1 or an integer >= 1 for --n-jobs, got {value}.")


def parse_finite_float(raw: str, *, flag: str) -> float:
    """argparse helper: parse a finite float."""

    try:
        value = float(raw)
    except ValueError as exc:
        raise argparse.ArgumentTypeError(
            f"Invalid {flag} value '{raw}'. Expected a number."
        ) from exc
    if not math.isfinite(value):
        raise argparse.ArgumentTypeError(f"Invalid {flag} value '{raw}'. Expected a finite number.")
    return value


def parse_bounded_float(
    raw: str,
    *,
    flag: str,
    lo: float,
    hi: float | None,
    lo_inclusive: bool,
    hi_inclusive: bool,
    expectation: str,
) -> float:
    """argparse helper: parse a finite float and enforce explicit numeric bounds."""

    value = parse_finite_float(raw, flag=flag)
    lo_ok = value >= lo if lo_inclusive else value > lo
    hi_ok = True
    if hi is not None:
        hi_ok = value <= hi if hi_inclusive else value < hi
    if lo_ok and hi_ok:
        return value
    raise argparse.ArgumentTypeError(f"Invalid {flag} value '{raw}'. Expected {expectation}.")


def parse_missing_rate_arg(raw: str) -> float:
    """argparse type: parse missing rate in [0, 1]."""

    return parse_bounded_float(
        raw,
        flag="--missing-rate",
        lo=0.0,
        hi=1.0,
        lo_inclusive=True,
        hi_inclusive=True,
        expectation="a finite value in [0, 1]",
    )


def parse_missing_mar_observed_fraction_arg(raw: str) -> float:
    """argparse type: parse MAR observed-feature fraction in (0, 1]."""

    return parse_bounded_float(
        raw,
        flag="--missing-mar-observed-fraction",
        lo=0.0,
        hi=1.0,
        lo_inclusive=False,
        hi_inclusive=True,
        expectation="a finite value in (0, 1]",
    )


def parse_missing_mar_logit_scale_arg(raw: str) -> float:
    """argparse type: parse MAR logit scale > 0."""

    return parse_bounded_float(
        raw,
        flag="--missing-mar-logit-scale",
        lo=0.0,
        hi=None,
        lo_inclusive=False,
        hi_inclusive=False,
        expectation="a finite value > 0",
    )


def parse_missing_mnar_logit_scale_arg(raw: str) -> float:
    """argparse type: parse MNAR logit scale > 0."""

    return parse_bounded_float(
        raw,
        flag="--missing-mnar-logit-scale",
        lo=0.0,
        hi=None,
        lo_inclusive=False,
        hi_inclusive=False,
        expectation="a finite value > 0",
    )


def parse_thresholds_csv_arg(raw: str) -> list[float]:
    """argparse type: parse threshold sweep CSV values in [0, 1]."""

    parts = [part.strip() for part in str(raw).split(",")]
    if not parts or any(part == "" for part in parts):
        raise argparse.ArgumentTypeError(
            "Invalid --thresholds value. Expected a CSV list of finite values in [0.0, 1.0]."
        )
    return [
        parse_bounded_float(
            part,
            flag="--thresholds",
            lo=FILTER_THRESHOLD_MIN,
            hi=FILTER_THRESHOLD_MAX,
            lo_inclusive=True,
            hi_inclusive=True,
            expectation=FILTER_THRESHOLD_EXPECTATION,
        )
        for part in parts
    ]


def parse_filter_threshold_arg(raw: str) -> float:
    """argparse type: parse one deferred-filter threshold in [0, 1]."""

    return parse_bounded_float(
        raw,
        flag="--threshold",
        lo=FILTER_THRESHOLD_MIN,
        hi=FILTER_THRESHOLD_MAX,
        lo_inclusive=True,
        hi_inclusive=True,
        expectation=FILTER_THRESHOLD_EXPECTATION,
    )


def parse_warn_threshold_pct_arg(raw: str) -> float:
    """argparse type: parse non-negative finite warn threshold percentages."""

    return parse_bounded_float(
        raw,
        flag="--warn-threshold-pct",
        lo=0.0,
        hi=None,
        lo_inclusive=True,
        hi_inclusive=False,
        expectation="a finite value >= 0",
    )


def parse_fail_threshold_pct_arg(raw: str) -> float:
    """argparse type: parse non-negative finite fail threshold percentages."""

    return parse_bounded_float(
        raw,
        flag="--fail-threshold-pct",
        lo=0.0,
        hi=None,
        lo_inclusive=True,
        hi_inclusive=False,
        expectation="a finite value >= 0",
    )


def parse_missing_mechanism_arg(raw: str) -> str:
    """argparse type: normalize missingness mechanism values."""

    try:
        return normalize_missing_mechanism(raw)
    except ValueError as exc:
        raise argparse.ArgumentTypeError(str(exc)) from exc



from __future__ import annotations

import numpy as np
import pytest

from dagzoo.io.lineage_artifact import (
    pack_upper_triangle_adjacency,
    sha256_hex,
    unpack_upper_triangle_adjacency,
    upper_triangle_bit_length,
)


def test_pack_unpack_upper_triangle_roundtrip() -> None:
    adjacency = [
        [0, 1, 1, 0],
        [0, 0, 1, 1],
        [0, 0, 0, 1],
        [0, 0, 0, 0],
    ]
    n_nodes, edge_count, payload = pack_upper_triangle_adjacency(adjacency)

    assert n_nodes == 4
    assert edge_count == 5
    assert len(payload) == 1

    restored = unpack_upper_triangle_adjacency(
        payload,
        n_nodes=n_nodes,
        bit_length=upper_triangle_bit_length(n_nodes),
    )
    assert restored.tolist() == adjacency


def test_pack_upper_triangle_rejects_lower_triangle_edges() -> None:
    adjacency = [
        [0, 1, 0],
        [1, 0, 1],
        [0, 0, 0],
    ]
    with pytest.raises(ValueError, match="upper triangular"):
        pack_upper_triangle_adjacency(adjacency)


def test_pack_upper_triangle_rejects_non_integral_values() -> None:
    adjacency = np.array(
        [
            [0.0, 0.9, 0.0],
            [0.0, 0.0, 1.2],
            [0.0, 0.0, 0.0],
        ],
        dtype=np.float64,
    )
    with pytest.raises(ValueError, match="entries must be 0 or 1"):
        pack_upper_triangle_adjacency(adjacency)


def test_unpack_upper_triangle_rejects_wrong_payload_length() -> None:
    with pytest.raises(ValueError, match="payload length does not match expected byte length"):
        unpack_upper_triangle_adjacency(b"", n_nodes=4, bit_length=6)


def test_sha256_hex_returns_expected_length() -> None:
    digest = sha256_hex(b"abc")
    assert len(digest) == 64



"""Storage helpers."""

from .lineage_artifact import (
    pack_upper_triangle_adjacency,
    resolve_lineage_path,
    sha256_hex,
    unpack_upper_triangle_adjacency,
    upper_triangle_bit_length,
)
from .lineage_schema import (
    LINEAGE_ADJACENCY_ENCODING,
    LINEAGE_SCHEMA_NAME,
    LINEAGE_SCHEMA_VERSION,
    LINEAGE_SCHEMA_VERSION_COMPACT,
    LINEAGE_SCHEMA_VERSION_DENSE,
    LineageValidationError,
    validate_lineage_payload,
    validate_metadata_lineage,
)
from .parquet_writer import write_packed_parquet_shards_stream

__all__ = [
    "LINEAGE_ADJACENCY_ENCODING",
    "LINEAGE_SCHEMA_NAME",
    "LINEAGE_SCHEMA_VERSION",
    "LINEAGE_SCHEMA_VERSION_COMPACT",
    "LINEAGE_SCHEMA_VERSION_DENSE",
    "LineageValidationError",
    "pack_upper_triangle_adjacency",
    "resolve_lineage_path",
    "sha256_hex",
    "unpack_upper_triangle_adjacency",
    "upper_triangle_bit_length",
    "validate_lineage_payload",
    "validate_metadata_lineage",
    "write_packed_parquet_shards_stream",
]



"""Benchmark guardrail helpers for lineage and missingness checks."""

from __future__ import annotations

import math
import pickle
import statistics
import tempfile
import time
from collections.abc import Iterable
from pathlib import Path
from typing import Any

from dagzoo.bench.constants import (
    LINEAGE_GUARDRAIL_RUNTIME_GATING_MIN_SAMPLE,
    LINEAGE_GUARDRAIL_RUNTIME_GATING_MIN_SAMPLE_SMOKE,
    LINEAGE_GUARDRAIL_RUNTIME_TRIALS,
    LINEAGE_GUARDRAIL_SMOKE_SAMPLE_CAP,
    SECONDS_PER_MINUTE,
)
from dagzoo.bench.metrics import degradation_percent
from dagzoo.config import GeneratorConfig
from dagzoo.core.dataset import generate_batch_iter
from dagzoo.io.parquet_writer import write_packed_parquet_shards_stream
from dagzoo.math import to_numpy as _to_numpy
from dagzoo.rng import KeyedRng
from dagzoo.types import DatasetBundle


def _lineage_guardrail_sample_count(config: GeneratorConfig, suite: str, num_datasets: int) -> int:
    """Choose dataset count used for lineage overhead guardrail checks."""

    n = max(1, min(int(config.benchmark.latency_num_samples), num_datasets))
    if suite == "smoke":
        return min(n, LINEAGE_GUARDRAIL_SMOKE_SAMPLE_CAP)
    return n


def _bundle_without_lineage(bundle: DatasetBundle) -> DatasetBundle:
    """Clone one bundle while removing lineage metadata for control persistence runs."""

    metadata = dict(bundle.metadata)
    metadata.pop("lineage", None)
    return DatasetBundle(
        X_train=bundle.X_train,
        y_train=bundle.y_train,
        X_test=bundle.X_test,
        y_test=bundle.y_test,
        feature_types=list(bundle.feature_types),
        metadata=metadata,
    )


def _measure_persistence_datasets_per_minute(
    bundles: Iterable[DatasetBundle],
    *,
    config: GeneratorConfig,
    num_bundles: int,
) -> float:
    """Measure write-path throughput for a fixed number of streamed bundles."""

    if num_bundles <= 0:
        return 0.0

    start = time.perf_counter()
    with tempfile.TemporaryDirectory(prefix="dagzoo_lineage_guardrail_") as tmp_dir:
        out_dir = Path(tmp_dir) / "shards"
        _ = write_packed_parquet_shards_stream(
            bundles,
            out_dir=out_dir,
            shard_size=max(1, int(config.output.shard_size)),
            compression=str(config.output.compression),
        )
    elapsed = time.perf_counter() - start
    if elapsed <= 0.0:
        return 0.0
    return (float(num_bundles) / elapsed) * SECONDS_PER_MINUTE


def _stage_bundle(path: Path, bundle: DatasetBundle, *, strip_lineage: bool) -> None:
    """Serialize one bundle to disk for later replay in persistence timing."""

    staged_bundle = _bundle_without_lineage(bundle) if strip_lineage else bundle
    payload: dict[str, Any] = {
        "X_train": _to_numpy(staged_bundle.X_train),
        "y_train": _to_numpy(staged_bundle.y_train),
        "X_test": _to_numpy(staged_bundle.X_test),
        "y_test": _to_numpy(staged_bundle.y_test),
        "feature_types": list(staged_bundle.feature_types),
        "metadata": dict(staged_bundle.metadata),
    }
    with path.open("wb") as f:
        pickle.dump(payload, f, protocol=pickle.HIGHEST_PROTOCOL)


def _iter_staged_bundles(stage_dir: Path, *, num_bundles: int) -> Iterable[DatasetBundle]:
    """Replay staged bundles from disk as a streaming iterator."""

    for idx in range(max(0, num_bundles)):
        bundle_path = stage_dir / f"bundle_{idx:06d}.pkl"
        with bundle_path.open("rb") as f:
            payload = pickle.load(f)
        if not isinstance(payload, dict):
            raise ValueError(f"Invalid staged bundle payload at {bundle_path}.")

        feature_types_raw = payload.get("feature_types")
        metadata_raw = payload.get("metadata")
        if not isinstance(feature_types_raw, list):
            raise ValueError(f"Staged feature_types must be a list at {bundle_path}.")
        if not isinstance(metadata_raw, dict):
            raise ValueError(f"Staged metadata must be an object at {bundle_path}.")
        feature_types = [str(value) for value in feature_types_raw]
        metadata = dict(metadata_raw)

        yield DatasetBundle(
            X_train=payload["X_train"],
            y_train=payload["y_train"],
            X_test=payload["X_test"],
            y_test=payload["y_test"],
            feature_types=feature_types,
            metadata=metadata,
        )


def _stage_lineage_trial_bundles(
    config: GeneratorConfig,
    *,
    sample_n: int,
    seed: int,
    device: str | None,
    baseline_stage_dir: Path,
    current_stage_dir: Path,
) -> tuple[int, int]:
    """Generate and stage baseline/current trial bundles once outside timed sections."""

    baseline_stage_dir.mkdir(parents=True, exist_ok=True)
    current_stage_dir.mkdir(parents=True, exist_ok=True)
    seen = 0
    with_lineage = 0
    bundle_iter = generate_batch_iter(
        config,
        num_datasets=sample_n,
        seed=seed,
        device=device,
    )
    for idx, bundle in enumerate(bundle_iter):
        has_lineage = isinstance(bundle.metadata.get("lineage"), dict)
        if has_lineage:
            with_lineage += 1
        file_name = f"bundle_{idx:06d}.pkl"
        _stage_bundle(current_stage_dir / file_name, bundle, strip_lineage=False)
        _stage_bundle(baseline_stage_dir / file_name, bundle, strip_lineage=True)
        seen = idx + 1
    return with_lineage, seen


def _measure_lineage_persistence_trials(
    *,
    baseline_stage_dir: Path,
    current_stage_dir: Path,
    num_bundles: int,
    config: GeneratorConfig,
    trials: int,
) -> tuple[list[float], list[float]]:
    """Collect paired baseline/lineage persistence throughput trials."""

    n_trials = max(1, int(trials))
    baseline_trials: list[float] = []
    current_trials: list[float] = []
    for _ in range(n_trials):
        baseline_bundles = _iter_staged_bundles(baseline_stage_dir, num_bundles=num_bundles)
        baseline_trials.append(
            _measure_persistence_datasets_per_minute(
                baseline_bundles,
                config=config,
                num_bundles=num_bundles,
            )
        )
        current_bundles = _iter_staged_bundles(current_stage_dir, num_bundles=num_bundles)
        current_trials.append(
            _measure_persistence_datasets_per_minute(
                current_bundles,
                config=config,
                num_bundles=num_bundles,
            )
        )
    return baseline_trials, current_trials


def _median_throughput(values: list[float]) -> float:
    """Return robust median throughput across trial values."""

    finite = [float(value) for value in values if math.isfinite(float(value))]
    if not finite:
        return 0.0
    return float(statistics.median(finite))


def _build_guardrail_issue(
    *,
    metric: str,
    severity: str,
    current: float | None,
    baseline: float | None,
    degradation_pct: float | None,
    detail: str,
) -> dict[str, Any]:
    """Create a normalized guardrail issue payload."""

    return {
        "metric": metric,
        "severity": severity,
        "current": current,
        "baseline": baseline,
        "degradation_pct": degradation_pct,
        "detail": detail,
    }


def _severity_from_thresholds(value: float, *, warn: float, fail: float) -> str:
    """Map a numeric value to pass/warn/fail severity."""

    if value >= fail:
        return "fail"
    if value >= warn:
        return "warn"
    return "pass"


def _status_from_issues(issues: list[dict[str, Any]]) -> str:
    """Collapse per-metric issues into an overall status."""

    severities = {str(issue.get("severity", "pass")) for issue in issues}
    if "fail" in severities:
        return "fail"
    if "warn" in severities:
        return "warn"
    return "pass"


def _issue_sort_key(issue: dict[str, Any]) -> tuple[int, float]:
    """Sort regression issues by severity then descending degradation percentage."""

    severity = str(issue.get("severity", "warn"))
    rank = 0 if severity == "fail" else 1 if severity == "warn" else 2
    raw_degradation = issue.get("degradation_pct")
    if isinstance(raw_degradation, bool) or not isinstance(raw_degradation, (int, float)):
        return (rank, 0.0)
    degradation = float(raw_degradation)
    if not math.isfinite(degradation):
        return (rank, 0.0)
    return (rank, -degradation)


def _collect_guardrail_regression_issues(
    preset_results: list[dict[str, Any]],
    *,
    guardrail_key: str,
) -> list[dict[str, Any]]:
    """Flatten one guardrail type into regression issue payloads."""

    issues: list[dict[str, Any]] = []
    for result in preset_results:
        guardrails = result.get(guardrail_key)
        if not isinstance(guardrails, dict) or not bool(guardrails.get("enabled")):
            continue
        raw_issues = guardrails.get("issues")
        if not isinstance(raw_issues, list):
            continue
        for issue in raw_issues:
            if not isinstance(issue, dict):
                continue
            merged = dict(issue)
            merged["preset"] = str(result.get("preset_key"))
            issues.append(merged)
    return issues


def _collect_lineage_guardrails(
    config: GeneratorConfig,
    *,
    suite: str,
    num_datasets: int,
    device: str | None,
    warn_threshold_pct: float,
    fail_threshold_pct: float,
) -> dict[str, Any]:
    """Measure lineage export overhead and convert it into guardrail metrics."""
    sample_n = _lineage_guardrail_sample_count(config, suite=suite, num_datasets=num_datasets)
    if sample_n <= 0:
        return {"enabled": False}

    sample_seed = KeyedRng(int(config.seed)).child_seed("bench", "guardrails", "lineage", "sample")
    with tempfile.TemporaryDirectory(prefix="dagzoo_lineage_guardrail_stage_") as tmp_dir:
        stage_root = Path(tmp_dir)
        baseline_stage_dir = stage_root / "baseline"
        current_stage_dir = stage_root / "current"
        try:
            bundles_with_lineage, bundles_seen = _stage_lineage_trial_bundles(
                config,
                sample_n=sample_n,
                seed=sample_seed,
                device=device,
                baseline_stage_dir=baseline_stage_dir,
                current_stage_dir=current_stage_dir,
            )
        except Exception as exc:
            return {
                "enabled": False,
                "reason": "sampling_failed",
                "detail": str(exc),
            }
        if bundles_seen <= 0:
            return {"enabled": False}

        lineage_coverage_rate = float(bundles_with_lineage) / float(bundles_seen)

        try:
            baseline_trials, current_trials = _measure_lineage_persistence_trials(
                baseline_stage_dir=baseline_stage_dir,
                current_stage_dir=current_stage_dir,
                num_bundles=bundles_seen,
                config=config,
                trials=LINEAGE_GUARDRAIL_RUNTIME_TRIALS,
            )
        except Exception as exc:
            return {
                "enabled": False,
                "reason": "unavailable",
                "detail": str(exc),
            }
    baseline_dpm = _median_throughput(baseline_trials)
    current_dpm = _median_throughput(current_trials)

    runtime_degradation = degradation_percent("datasets_per_minute", current_dpm, baseline_dpm)
    runtime_degradation_value = (
        float(runtime_degradation) if runtime_degradation is not None else 0.0
    )
    runtime_severity = _severity_from_thresholds(
        runtime_degradation_value,
        warn=float(warn_threshold_pct),
        fail=float(fail_threshold_pct),
    )
    runtime_gating_min_sample = (
        LINEAGE_GUARDRAIL_RUNTIME_GATING_MIN_SAMPLE_SMOKE
        if suite == "smoke"
        else LINEAGE_GUARDRAIL_RUNTIME_GATING_MIN_SAMPLE
    )
    runtime_gating_enabled = sample_n >= runtime_gating_min_sample

    issues: list[dict[str, Any]] = []
    if bundles_with_lineage != bundles_seen:
        issues.append(
            _build_guardrail_issue(
                metric="lineage_metadata_coverage",
                severity="fail",
                current=float(lineage_coverage_rate),
                baseline=1.0,
                degradation_pct=float(max(0.0, (1.0 - lineage_coverage_rate) * 100.0)),
                detail="Lineage metadata must be present for all generated bundles.",
            )
        )
    if runtime_gating_enabled and runtime_severity != "pass":
        issues.append(
            _build_guardrail_issue(
                metric="lineage_export_runtime_degradation_pct",
                severity=runtime_severity,
                current=float(current_dpm),
                baseline=float(baseline_dpm),
                degradation_pct=float(runtime_degradation_value),
                detail=(
                    "Lineage-enabled shard persistence throughput regressed versus a "
                    "lineage-stripped control persistence run."
                ),
            )
        )

    return {
        "enabled": True,
        "sample_datasets": int(sample_n),
        "runtime_gating_enabled": bool(runtime_gating_enabled),
        "runtime_gating_min_sample_datasets": int(runtime_gating_min_sample),
        "runtime_gating_suppressed_reason": (
            None if runtime_gating_enabled else "insufficient_sample_size"
        ),
        "runtime_trials": int(max(1, LINEAGE_GUARDRAIL_RUNTIME_TRIALS)),
        "runtime_baseline_trials_dpm": baseline_trials,
        "runtime_with_lineage_trials_dpm": current_trials,
        "lineage_metadata_coverage_rate": float(lineage_coverage_rate),
        "runtime_baseline_datasets_per_minute": float(baseline_dpm),
        "runtime_with_lineage_datasets_per_minute": float(current_dpm),
        "runtime_degradation_pct": (
            float(runtime_degradation) if runtime_degradation is not None else None
        ),
        "runtime_warn_threshold_pct": float(warn_threshold_pct),
        "runtime_fail_threshold_pct": float(fail_threshold_pct),
        "issues": issues,
        "status": _status_from_issues(issues),
    }



"""Dataset split validation helpers."""

from __future__ import annotations

import math

import torch


class InfeasibleStratifiedSplitError(ValueError):
    """Raised when a classification split cannot satisfy stratification constraints."""


class InvalidClassSplitError(ValueError):
    """Raised when a classification split violates emitted bundle invariants."""


def _stratified_split_indices(
    y: torch.Tensor,
    n_train: int,
    generator: torch.Generator,
    device: str,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Return (train_indices, test_indices) with proportional class representation.

    For classification tasks this keeps class balance close to proportional and
    ensures classes with at least two members appear in both splits. For
    infeasible combinations, this raises ``InfeasibleStratifiedSplitError``
    with an ``infeasible_stratified_split`` prefix.
    """
    n_total = int(y.shape[0])
    n_test = n_total - n_train
    if n_total <= 0 or n_train <= 0 or n_test <= 0:
        raise InfeasibleStratifiedSplitError(
            f"infeasible_stratified_split: expected 0 < n_train < n_total, got n_train={n_train}, n_total={n_total}."
        )

    classes = torch.unique(y, sorted=True)
    train_frac = n_train / n_total

    cls_indices: list[torch.Tensor] = []
    cls_values: list[int] = []
    cls_train_counts: list[int] = []
    cls_train_min: list[int] = []
    cls_train_max: list[int] = []
    cls_remainders: list[float] = []

    for cls in classes:
        idx = torch.where(y == cls)[0]
        perm = torch.randperm(idx.shape[0], generator=generator, device=device)
        cls_indices.append(idx[perm])
        cls_values.append(int(cls.item()))

        n_cls = int(idx.shape[0])
        proportional = float(n_cls * train_frac)
        base_alloc = int(math.floor(proportional))
        remainder = proportional - base_alloc
        if n_cls >= 2:
            train_min = 1
            train_max = n_cls - 1
        else:
            train_min = 0
            train_max = n_cls

        n_cls_train = max(train_min, min(base_alloc, train_max))
        cls_train_counts.append(n_cls_train)
        cls_train_min.append(train_min)
        cls_train_max.append(train_max)
        cls_remainders.append(remainder)

    deficit = n_train - sum(cls_train_counts)
    if deficit > 0:
        order = sorted(
            range(len(cls_train_counts)), key=lambda i: (-cls_remainders[i], cls_values[i])
        )
        while deficit > 0:
            progressed = False
            for i in order:
                if cls_train_counts[i] < cls_train_max[i]:
                    cls_train_counts[i] += 1
                    deficit -= 1
                    progressed = True
                    if deficit == 0:
                        break
            if not progressed:
                break
        if deficit > 0:
            raise InfeasibleStratifiedSplitError(
                "infeasible_stratified_split: unable to allocate requested train rows while "
                f"preserving class constraints (remaining={deficit})."
            )
    elif deficit < 0:
        surplus = -deficit
        order = sorted(
            range(len(cls_train_counts)), key=lambda i: (cls_remainders[i], cls_values[i])
        )
        while surplus > 0:
            progressed = False
            for i in order:
                if cls_train_counts[i] > cls_train_min[i]:
                    cls_train_counts[i] -= 1
                    surplus -= 1
                    progressed = True
                    if surplus == 0:
                        break
            if not progressed:
                break
        if surplus > 0:
            raise InfeasibleStratifiedSplitError(
                "infeasible_stratified_split: unable to allocate requested test rows while "
                f"preserving class constraints (remaining={surplus})."
            )

    if sum(cls_train_counts) != n_train:
        raise InfeasibleStratifiedSplitError(
            "infeasible_stratified_split: train allocation mismatch after rebalance "
            f"(expected={n_train}, actual={sum(cls_train_counts)})."
        )

    train_parts: list[torch.Tensor] = []
    test_parts: list[torch.Tensor] = []
    for idx, n_cls_train in zip(cls_indices, cls_train_counts, strict=True):
        train_parts.append(idx[:n_cls_train])
        test_parts.append(idx[n_cls_train:])

    train_idx = torch.cat(train_parts)
    test_idx = torch.cat(test_parts)
    if int(train_idx.shape[0]) != n_train or int(test_idx.shape[0]) != n_test:
        raise InfeasibleStratifiedSplitError(
            "infeasible_stratified_split: index cardinality mismatch "
            f"(expected_train={n_train}, actual_train={int(train_idx.shape[0])}, "
            f"expected_test={n_test}, actual_test={int(test_idx.shape[0])})."
        )

    # Shuffle within each split
    train_idx = train_idx[torch.randperm(train_idx.shape[0], generator=generator, device=device)]
    test_idx = test_idx[torch.randperm(test_idx.shape[0], generator=generator, device=device)]

    return train_idx, test_idx


def _classification_split_valid(y_train: torch.Tensor, y_test: torch.Tensor) -> bool:
    """Validate classification split constraints."""

    train_classes = torch.unique(y_train.to(torch.int64), sorted=True)
    test_classes = torch.unique(y_test.to(torch.int64), sorted=True)
    return bool(train_classes.numel() >= 2 and torch.equal(train_classes, test_classes))



from __future__ import annotations

import json
from pathlib import Path

from conftest import load_script_module


def _write_json(path: Path, payload: object) -> Path:
    path.write_text(json.dumps(payload), encoding="utf-8")
    return path


def test_build_and_render_benchmark_pr_report(tmp_path) -> None:
    module = load_script_module("benchmark_pr_report", "scripts/ci/benchmark_pr_report.py")
    summary_path = tmp_path / "summary.json"
    baseline_path = tmp_path / "baseline.json"

    summary_payload = {
        "suite": "smoke",
        "preset_results": [
            {
                "preset_key": "cpu",
                "datasets_per_minute": 85.0,
                "generation_datasets_per_minute": 100.0,
                "write_datasets_per_minute": 100.0,
                "filter_datasets_per_minute": None,
                "filter_accepted_datasets_per_minute": None,
            }
        ],
    }
    baseline_payload = {
        "version": 1,
        "suite": "smoke",
        "metrics": [
            "datasets_per_minute",
            "generation_datasets_per_minute",
            "write_datasets_per_minute",
        ],
        "presets": {
            "cpu": {
                "datasets_per_minute": 100.0,
                "generation_datasets_per_minute": 100.0,
                "write_datasets_per_minute": 100.0,
            }
        },
    }
    _write_json(summary_path, summary_payload)
    _write_json(baseline_path, baseline_payload)

    report = module.build_benchmark_pr_report(
        summary_payload,
        baseline_payload,
        title="Benchmark Smoke (CPU)",
        summary_path=summary_path,
        baseline_path=baseline_path,
        warn_threshold_pct=10.0,
        fail_threshold_pct=20.0,
    )
    comment = module.render_comment_markdown(report)
    step_summary = module.render_step_summary_markdown(
        {
            **report,
            "comment_path": "benchmarks/results/ci_smoke/benchmark_pr_comment.md",
            "delta_path": "benchmarks/results/ci_smoke/benchmark_delta.json",
        }
    )
    comment_path, step_summary_path, delta_path = module.write_report_artifacts(
        {
            **report,
            "comment_path": "benchmarks/results/ci_smoke/benchmark_pr_comment.md",
            "delta_path": "benchmarks/results/ci_smoke/benchmark_delta.json",
        },
        comment_out=tmp_path / "benchmark_pr_comment.md",
        step_summary_out=tmp_path / "benchmark_pr_step_summary.md",
        delta_out=tmp_path / "benchmark_delta.json",
    )

    assert report["status"] == "warn"
    assert report["issue_count"] == 1
    assert "| Metric | Current | Baseline | Delta % | Severity |" in comment
    assert "datasets_per_minute" in comment
    assert "warn" in comment
    assert "## Artifacts" in step_summary
    assert comment_path.exists()
    assert step_summary_path.exists()
    assert delta_path.exists()
    delta_payload = json.loads(delta_path.read_text(encoding="utf-8"))
    assert delta_payload["status"] == "warn"
    assert delta_payload["preset_reports"][0]["rows"][0]["severity"] == "warn"


def test_main_handles_missing_summary_file(tmp_path) -> None:
    module = load_script_module("benchmark_pr_report_missing", "scripts/ci/benchmark_pr_report.py")
    baseline_path = tmp_path / "baseline.json"
    _write_json(
        baseline_path,
        {
            "version": 1,
            "suite": "smoke",
            "metrics": ["datasets_per_minute"],
            "presets": {"cpu": {"datasets_per_minute": 100.0}},
        },
    )

    exit_code = module.main(
        [
            "--summary-json",
            str(tmp_path / "missing_summary.json"),
            "--baseline-json",
            str(baseline_path),
            "--title",
            "Benchmark Smoke (CPU)",
            "--comment-out",
            str(tmp_path / "comment.md"),
            "--step-summary-out",
            str(tmp_path / "step_summary.md"),
            "--delta-out",
            str(tmp_path / "delta.json"),
        ]
    )

    assert exit_code == 0
    assert "Benchmark summary file was not generated." in (tmp_path / "comment.md").read_text(
        encoding="utf-8"
    )
    assert '"missing-summary"' in (tmp_path / "delta.json").read_text(encoding="utf-8")
